import functools as ft
from typing import Any

import jax
import jax.core
import jax.extend.core
import jax.interpreters.ad as ad
import jax.interpreters.batching as batching
import jax.interpreters.mlir as mlir
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import PyTree

from .._filters import combine, is_array, is_array_like, partition
from .._tree import tree_equal


#
# filter primitives
# -----------------
# As with all filtering in Equinox, this is basically just about putting a
# nicer interface on existing JAX operations; in this case creating custom
# primitives. The inputs and outputs to the primitive can be arbitrary.
#


_like_sentinel = object()
_dummy_none = object()
_missing_dynamic = object()


# Not a PyTree
class _WrappedPrimal:
    def __init__(self, value):
        self.value = value


def _wrap_undefined(x):
    if isinstance(x, ad.UndefinedPrimal):
        return _WrappedPrimal(x)
    else:
        return x


def _unwrap_undefined(x, aval=False):
    if isinstance(x, _WrappedPrimal):
        if aval:
            return x.value.aval
        else:
            return x.value
    else:
        return x


def _is_array_like_internal(x):
    assert type(x) is not ad.UndefinedPrimal
    # Not `type(x) in (...)` as that doesn't handle stuff like ConcreteArrays.
    return is_array(x) or isinstance(x, (jax.core.ShapedArray, _WrappedPrimal))


def _zero_from_primal(p):
    assert type(p) is not ad.UndefinedPrimal
    if hasattr(jax, "typeof"):
        aval = jax.typeof(p)
    else:
        aval = jax.core.get_aval(p)
    if hasattr(aval, "to_tangent_aval"):
        # JAX >=0.4.34
        aval = aval.to_tangent_aval()  # pyright: ignore
    else:
        # earlier JAX
        aval = aval.at_least_vspace()
    return ad.Zero(aval)


def _combine(dynamic, static):
    iter_dynamic = iter(dynamic)
    out = [next(iter_dynamic) if x is _missing_dynamic else x for x in static]
    assert next(iter_dynamic, None) is None
    return out


def _is_none(x):
    return x is None


def _replace_none(x):
    return _dummy_none if x is None else x


def _get_second(x, y):
    return None if x is None else y


def _make_spec(x, y):
    if y is None:
        return _is_array_like_internal(x)
    else:
        assert x is not None
        return True


class Flatten:
    __slots__ = ("treedef_out", "static_out", "__weakref__")

    def called(self):
        return hasattr(self, "treedef_out")

    def get(self):
        return self.treedef_out, self.static_out

    def __call__(self, out, like=_like_sentinel):
        if like is _like_sentinel:
            dynamic_out, static_out = partition(out, _is_array_like_internal)
            flat_out, treedef_out = jtu.tree_flatten(dynamic_out)
            try:
                treedef_out_old = self.treedef_out
                static_out_old = self.static_out
            except AttributeError:
                self.treedef_out = treedef_out
                self.static_out = static_out
            else:
                assert treedef_out_old == treedef_out
                assert tree_equal(static_out_old, static_out)
            return flat_out
        else:
            assert jtu.tree_structure(out, is_leaf=_is_none) == jtu.tree_structure(
                like, is_leaf=_is_none
            )
            spec = jtu.tree_map(_make_spec, out, like, is_leaf=_is_none)
            dynamic_out, static_out = partition(out, spec, is_leaf=_is_none)
            flat_out, treedef_out = jtu.tree_flatten(dynamic_out)
            try:
                treedef_out_old = self.treedef_out
                static_out_old = self.static_out
            except AttributeError:
                self.treedef_out = treedef_out
                self.static_out = static_out
            else:
                assert treedef_out_old == treedef_out
                assert tree_equal(static_out_old, static_out)
            like = jtu.tree_map(_replace_none, like, is_leaf=_is_none)
            like = jtu.tree_map(_get_second, dynamic_out, like, is_leaf=_is_none)
            flat_like, treedef_like = jtu.tree_flatten(like)
            flat_like = [None if x is _dummy_none else x for x in flat_like]
            assert treedef_like == treedef_out
            assert len(flat_out) == len(flat_like)
            return flat_out, flat_like


def filter_primitive_def(rule):
    """For wrapping def_impl and def_abstract_eval.

    These can now take arbitrary inputs and outputs.
    """

    def _wrapper(*dynamic, treedef, static, flatten):
        args = jtu.tree_unflatten(treedef, _combine(dynamic, static))
        out = rule(*args)
        return flatten(out)

    return _wrapper


def filter_primitive_jvp(rule):
    """
    The input tangents (to the wrapped rule) will be a PyTree with the same
    structure as the input primals. `None` indicates symbolic zero tangents,
    in particular for non-JAX-array-likes.

    The output tangents are expected to match the output primals, necessarily
    with `None` for all non-JAX-array-likes.
    """

    def _wrapper(primals, tangents, *, treedef, static, flatten):
        tangents = [None if type(t) is ad.Zero else t for t in tangents]
        tangents_static = [x if x is _missing_dynamic else None for x in static]
        primals = jtu.tree_unflatten(treedef, _combine(primals, static))
        tangents = jtu.tree_unflatten(treedef, _combine(tangents, tangents_static))
        primals_out, tangents_out = rule(primals, tangents)
        flat_primals_out, flat_tangents_out = flatten(primals_out, tangents_out)
        flat_tangents_out = [
            _zero_from_primal(p) if t is None else t
            for p, t in zip(flat_primals_out, flat_tangents_out)
        ]
        return flat_primals_out, flat_tangents_out

    return _wrapper


_sentinel: Any = object()


def filter_primitive_transpose(rule=_sentinel, *, materialise_zeros=False):
    """
    The `inputs` to the transpose rule are a PyTree like the primal
    inputs, with `UndefinedPrimal`s where appropriate.

    The `cts_out` passed to the transpose rule are a PyTree like the
    primal output, with `None` for symbolic zero cotangents, in particular
    for non-JAX-array-likes.

    The output from the rule should be a PyTree like the primal input.
    All leaves which were non-JAX-array-like, or which should have zero
    cotangent, should have cotangent `None`.
    """
    if rule is _sentinel:
        return ft.partial(
            filter_primitive_transpose, materialise_zeros=materialise_zeros
        )

    def _wrapper(cts_out, *dynamic, treedef, static, flatten):
        treedef_out, _ = flatten.get()
        if materialise_zeros:
            cts_out = [ad.instantiate_zeros(ct) for ct in cts_out]
        else:
            cts_out = [None if type(ct) is ad.Zero else ct for ct in cts_out]
        cts_out = jtu.tree_unflatten(treedef_out, cts_out)
        wrapped_dynamic = [_wrap_undefined(x) for x in dynamic]
        wrapped_flat = _combine(wrapped_dynamic, static)
        wrapped_inputs = jtu.tree_unflatten(treedef, wrapped_flat)
        inputs = jtu.tree_map(_unwrap_undefined, wrapped_inputs)
        cts = rule(inputs, cts_out)
        flat_inputs, flat_cts = Flatten()(wrapped_inputs, cts)
        flat_inputs = [_unwrap_undefined(p, aval=True) for p in flat_inputs]
        flat_cts = [
            _zero_from_primal(p) if ct is None else ct
            for p, ct in zip(flat_inputs, flat_cts)
        ]
        assert len(dynamic) == len(flat_cts)
        return flat_cts

    return _wrapper


def filter_primitive_batching(rule):
    """
    The input batch axes (to the wrapped rule) will be a PyTree with the same
    structure as the input primals, with `None` for all non-JAX-arrays.

    The output batch axes are expected to match the output primals, with `None`
    for all non-JAX-arrays.
    """

    def _wrapper(dynamic, batch_axes, *, treedef, static, flatten):
        flat = _combine(dynamic, static)
        inputs = jtu.tree_unflatten(treedef, flat)
        batch_axes = [None if b is None else b for b in batch_axes]
        batch_axes_static = [x if x is _missing_dynamic else None for x in static]
        batch_axes = _combine(batch_axes, batch_axes_static)
        batch_axes = jtu.tree_unflatten(treedef, batch_axes)
        out, batch_axes = rule(inputs, batch_axes)
        flat_out, flat_batch_axes = flatten(out, batch_axes)
        flat_batch_axes = [None if b is None else b for b in flat_batch_axes]
        return flat_out, flat_batch_axes

    return _wrapper


def filter_primitive_bind(prim: jax.extend.core.Primitive, *args) -> PyTree:
    """Calls a primitive that has had its rules defined using the filter
    functions above.
    """
    assert prim.multiple_results
    # If `args` contains a Jaxpr or ClosedJaxpr in its leaves, then it ends up as a
    # member of the `static` tuple. This is important to ensure that jaxpr-rewriting
    # passes are able to find it.
    # (E.g. if `eqx.filter_closure_convert(...)` is an argument and we apply
    # `jax.core.jaxprs_in_params`.)
    flat, treedef = jtu.tree_flatten(args)
    dynamic = [x for x in flat if is_array(x)]
    static = tuple(_missing_dynamic if is_array(x) else x for x in flat)
    flatten = Flatten()
    flat_out = prim.bind(*dynamic, treedef=treedef, static=static, flatten=flatten)
    treedef_out, static_out = flatten.get()
    return combine(jtu.tree_unflatten(treedef_out, flat_out), static_out)


# Useful helper for JVP rules of higher-order primitives.
def materialise_zeros(primal, tangent, allow_struct=False):
    arraylike = is_array_like(primal)
    if allow_struct:
        arraylike = arraylike or isinstance(primal, jax.ShapeDtypeStruct)
    if tangent is None and arraylike:
        tangent = _zero_from_primal(primal)
        return ad.instantiate_zeros(tangent)
    else:
        return tangent


#
# vprim
# -----
# This allows for creating a primitive without needing to specify its batching rule.
# This is instead automatically obtained by vmap'ing its other rules (assuming that
# its rules are implemented using JAX operations).
#

_vprim_impl_registry = {}
_vprim_abstract_eval_registry = {}
_vprim_jvp_registry = {}
_vprim_transpose_registry = {}


def create_vprim(name: str, impl, abstract_eval, jvp, transpose):
    prim = jax.extend.core.Primitive(name)
    prim.multiple_results = True

    def batch_rule(axis_size, axis_name, trace_type, inputs, batch_axes, **params):
        del trace_type
        if all(b is None for b in jtu.tree_leaves(batch_axes)):
            out = prim.bind(*inputs, **params)
            batch_axes_out = jtu.tree_map(lambda _: None, out)
        else:
            # delegates batching to `_vprim_p`
            out = _vprim_p.bind(
                *inputs,
                prim=prim,
                __axis_size=axis_size,
                __axis_name=axis_name,
                __batch_axes=batch_axes,
                params=tuple(params.items()),
            )
            batch_axes_out = jtu.tree_map(lambda _: 0, out)
        return out, batch_axes_out

    prim.def_impl(impl)
    prim.def_abstract_eval(abstract_eval)
    ad.primitive_jvps[prim] = jvp
    ad.primitive_transposes[prim] = transpose
    batching.axis_primitive_batchers[prim] = batch_rule
    mlir.register_lowering(prim, mlir.lower_fun(impl, multiple_results=True))
    _vprim_impl_registry[prim] = impl
    _vprim_abstract_eval_registry[prim] = abstract_eval
    _vprim_jvp_registry[prim] = jvp
    _vprim_transpose_registry[prim] = transpose
    return prim


def _vprim_impl(*inputs, prim, __axis_size, __axis_name, __batch_axes, params):
    impl = ft.partial(_vprim_impl_registry[prim], **dict(params))
    impl = jax.vmap(
        impl, in_axes=__batch_axes, axis_size=__axis_size, axis_name=__axis_name
    )
    return impl(*inputs)


if hasattr(jax.extend.core, "mapped_aval"):
    _mapped_aval = jax.extend.core.mapped_aval  # pyright: ignore[reportAttributeAccessIssue]
else:
    _mapped_aval = jax.core.mapped_aval
if hasattr(jax.extend.core, "unmapped_aval"):
    _unmapped_aval = jax.extend.core.unmapped_aval  # pyright: ignore[reportAttributeAccessIssue,reportAssignmentType]
else:
    _unmapped_aval = jax.core.unmapped_aval  # pyright: ignore[reportAssignmentType]
if jax.__version_info__ >= (0, 5, 1):
    _old_unmapped_aval = _unmapped_aval

    def _unmapped_aval(axis_size, axis_name, axis, aval):
        del axis_name
        return _old_unmapped_aval(axis_size, axis, aval)  # pyright: ignore[reportCallIssue]


def _vprim_abstract_eval(*inputs, prim, __axis_size, __axis_name, __batch_axes, params):
    assert len(inputs) == len(__batch_axes)
    inputs = [_mapped_aval(__axis_size, b, x) for x, b in zip(inputs, __batch_axes)]
    abstract_eval = _vprim_abstract_eval_registry[prim]
    outs = abstract_eval(*inputs, **dict(params))
    outs = [_unmapped_aval(__axis_size, __axis_name, 0, x) for x in outs]
    return outs


def _resolve_zeros_t(tangent, batch_axis):
    if type(tangent) is ad.Zero and isinstance(batch_axis, int):
        aval = tangent.aval
        # Also accepts ConcreteArrays
        if not isinstance(aval, jax.core.ShapedArray):
            raise NotImplementedError(
                "vprim only currently supports ShapedArrays for symbolic zeros"
            )
        shape = aval.shape[:batch_axis] + aval.shape[batch_axis + 1 :]
        return ad.Zero(jax.core.ShapedArray(shape, aval.dtype))
    else:
        return tangent


def _resolve_zeros_b(tangent, batch_axis):
    if type(tangent) is ad.Zero:
        return None
    else:
        return batch_axis


def _vprim_jvp(
    primals, tangents, *, prim, __axis_size, __axis_name, __batch_axes, params
):
    assert len(primals) == len(__batch_axes)
    assert len(tangents) == len(__batch_axes)
    tangents = [_resolve_zeros_t(t, b) for t, b in zip(tangents, __batch_axes)]
    batch_axes_t = [_resolve_zeros_b(t, b) for t, b in zip(tangents, __batch_axes)]
    jvp = ft.partial(_vprim_jvp_registry[prim], **dict(params))
    jvp = jax.vmap(
        jvp,
        in_axes=(__batch_axes, batch_axes_t),
        axis_size=__axis_size,
        axis_name=__axis_name,
    )
    return jvp(primals, tangents)


def _resolve_undefined_i(input, batch_axis):
    if type(input) is ad.UndefinedPrimal and isinstance(batch_axis, int):
        aval = input.aval
        # Also accepts ConcreteArrays
        if not isinstance(aval, jax.core.ShapedArray):
            raise NotImplementedError(
                "vprim only currently supports ShapedArrays for undefined primals"
            )
        shape = aval.shape[:batch_axis] + aval.shape[batch_axis + 1 :]
        return ad.UndefinedPrimal(jax.core.ShapedArray(shape, aval.dtype))
    else:
        return input


def _resolve_undefined_b(input, batch_axis):
    if type(input) is ad.UndefinedPrimal:
        return None
    else:
        return batch_axis


def _vprim_transpose(
    cts, *inputs, prim, __axis_size, __axis_name, __batch_axes, params
):
    mapped_inputs = [_resolve_undefined_i(i, b) for i, b in zip(inputs, __batch_axes)]
    batch_axes = [_resolve_undefined_b(i, b) for i, b in zip(inputs, __batch_axes)]

    def _transpose(*_inputs):
        _outputs = _vprim_transpose_registry[prim](*_inputs, **dict(params))
        # `Zero` is not a JAX type -- it's an internal AD thing -- so we shouldn't pass
        # it across the `vmap` boundary. In particular JAX won't apply the out batch
        # axis to it.
        # JAX allows for returning `None` to indicate no cotangent, so we use that
        # instead, which is compatible with both `vmap` and `out_axes`.
        return tuple(None if type(o) is ad.Zero else o for o in _outputs)

    transpose = jax.vmap(
        _transpose,
        in_axes=(0, *batch_axes),
        out_axes=__batch_axes,
        axis_size=__axis_size,
        axis_name=__axis_name,
    )
    if prim.multiple_results:
        cts = tuple(None if type(c) is ad.Zero else c for c in cts)
    else:
        cts = None if type(cts) is ad.Zero else cts
    outputs = transpose(cts, *mapped_inputs)
    assert len(inputs) == len(outputs)
    for i, o in zip(inputs, outputs):
        if o is not None:
            # Can't have cotangents on defined variables I think? The point of an
            # `UndefinedPrimal` is to declare what you want cotangents with respect to.
            assert type(i) is ad.UndefinedPrimal
            # We've filtered out all other avals above, with a `NotImplementedError` if
            # required.
            assert isinstance(i.aval, jax.core.ShapedArray)
            assert i.aval.shape == jnp.shape(o)
    return outputs


# _vprim_p is itself a vprim!
_vprim_p = create_vprim(
    "vprim", _vprim_impl, _vprim_abstract_eval, _vprim_jvp, _vprim_transpose
)



import itertools as it
from typing import Any, TYPE_CHECKING

import jax
import jax.extend.core
import jax.interpreters.ad as ad
import jax.interpreters.batching as batching
import jax.interpreters.mlir as mlir
import jax.lax as lax
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import Array, Bool

from ..._filters import combine, is_array, partition
from ..._module import field, Module
from ..._pretty_print import tree_pformat
from ..._tree import tree_at, tree_equal
from ..._unvmap import unvmap_any
from .._nontraceable import nonbatchable
from .._primitive import create_vprim


def _select_if_vmap_impl(pred, x, y):
    # Not including the following, as it destroys performance on the GPU: it seems like
    # a copy of `x` is being made.
    #
    # msg = (
    #     "Internal error in Equinox. Please report a bug at "
    #     "https://github.com/patrick-kidger/equinox."
    # )
    # x = error_if(x, jnp.invert(pred), msg)
    return x


def _select_if_vmap_abstract(pred, x, y):
    return x


def _select_if_vmap_jvp(primals, tangents):
    pred, x, y = primals
    _, tx, ty = tangents
    assert x.shape == tx.aval.shape
    assert x.dtype == tx.aval.dtype
    assert y.shape == ty.aval.shape
    assert y.dtype == ty.aval.dtype
    out = _select_if_vmap(pred, x, y, makes_false_steps=False)
    if type(tx) is ad.Zero and type(ty) is ad.Zero:
        t_out = tx
    else:
        if type(tx) is ad.Zero:
            tx = jnp.zeros(tx.aval.shape, tx.aval.dtype)  # pyright: ignore
        if type(ty) is ad.Zero:
            ty = jnp.zeros(ty.aval.shape, ty.aval.dtype)  # pyright: ignore
        t_out = _select_if_vmap(pred, tx, ty, makes_false_steps=False)
    return out, t_out


def _select_if_vmap_transpose(ct, pred, x, y):
    assert not ad.is_undefined_primal(pred)
    assert ct.shape == x.aval.shape
    assert ct.dtype == x.aval.dtype
    assert ct.shape == y.aval.shape
    assert ct.dtype == y.aval.dtype
    if type(ct) is ad.Zero:
        ct_x = None
        ct_y = None
    else:
        zero = jnp.zeros(ct.shape, ct.dtype)
        if ad.is_undefined_primal(x):
            ct_x = _select_if_vmap(pred, ct, zero, makes_false_steps=False)
        else:
            ct_x = None
        if ad.is_undefined_primal(y):
            ct_y = _select_if_vmap(pred, zero, ct, makes_false_steps=False)
        else:
            ct_y = None
    return [None, ct_x, ct_y]


def _select_if_vmap_batch(axis_size, axis_name, trace, inputs, batch_axes):
    del axis_name, trace
    pred, x, y = inputs
    bp, bx, by = batch_axes
    if bp is None:
        if bx is None:
            if by is None:
                out_axis = None
            else:
                x = jnp.broadcast_to(x, (axis_size,) + x.shape)
                y = jnp.moveaxis(y, by, 0)
                out_axis = 0
        else:
            if by is None:
                x = jnp.moveaxis(x, bx, 0)
                y = jnp.broadcast_to(y, (axis_size,) + y.shape)
                out_axis = 0
            else:
                x = jnp.moveaxis(x, bx, 0)
                y = jnp.moveaxis(y, by, 0)
                out_axis = 0
        out = _select_if_vmap(pred, x, y, makes_false_steps=False)
    else:
        out = jax.vmap(lax.select, in_axes=(bp, bx, by))(pred, x, y)
        out_axis = 0
    return out, out_axis


select_if_vmap_p = jax.extend.core.Primitive("select_if_vmap")
select_if_vmap_p.def_impl(_select_if_vmap_impl)
select_if_vmap_p.def_abstract_eval(_select_if_vmap_abstract)
ad.primitive_jvps[select_if_vmap_p] = _select_if_vmap_jvp
ad.primitive_transposes[select_if_vmap_p] = _select_if_vmap_transpose
batching.axis_primitive_batchers[select_if_vmap_p] = _select_if_vmap_batch
mlir.register_lowering(
    select_if_vmap_p, mlir.lower_fun(_select_if_vmap_impl, multiple_results=False)
)


# We want to insert `lax.select`s to avoid updating any batch elements in the loop that
# have a False predicate. (But the loop is still going whilst other batch elements have
# a True predicate). However, if we have no vmap at all, then we can be slightly more
# efficient: don't introduce a select at all.
def _select_if_vmap(pred, x, y, makes_false_steps):
    """As `lax.select(pred, x, y)` if `pred` is vmap'd. Not-vmap'd `pred` are assumed to
    be `True`, so that in this case `x` is returned unconditionally.
    """
    if makes_false_steps:
        return lax.select(pred, x, y)
    else:
        pred = fixed_asarray(pred)
        assert pred.shape == ()
        assert pred.dtype == jnp.bool_
        x = fixed_asarray(x)
        y = fixed_asarray(y)
        assert x.shape == y.shape
        assert x.dtype == y.dtype
        return select_if_vmap_p.bind(pred, x, y)


def _maybe_set_impl(
    pred, xs, x, *i_dynamic_leaves, i_static, i_treedef, kwargs, makes_false_steps
):
    i_static = _to_raw_slice(i_static)
    i = combine(i_static, jtu.tree_unflatten(i_treedef, i_dynamic_leaves))
    kwargs_dict = dict(kwargs)
    x = _select_if_vmap(pred, x, xs.at[i].get(**kwargs_dict), makes_false_steps)
    return [xs.at[i].set(x, **kwargs_dict)]


def _maybe_set_abstract(
    pred, xs, x, *i_dynamic_leaves, i_static, i_treedef, kwargs, makes_false_steps
):
    del pred, i_dynamic_leaves, i_static, i_treedef, kwargs, makes_false_steps
    return [xs]


def _maybe_set_jvp(
    primals, tangents, *, i_static, i_treedef, kwargs, makes_false_steps
):
    pred, xs, x, *i_dynamic_leaves = primals
    _, t_xs, t_x, *_ = tangents
    i_static = _to_raw_slice(i_static)
    i = combine(i_static, jtu.tree_unflatten(i_treedef, i_dynamic_leaves))
    out = _maybe_set(
        pred, xs, x, i, kwargs=dict(kwargs), makes_false_steps=makes_false_steps
    )
    if type(t_x) is ad.Zero and type(t_xs) is ad.Zero:
        t_out = t_xs
    else:
        if type(t_x) is ad.Zero:
            t_x = jnp.zeros(t_x.aval.shape, t_x.aval.dtype)  # pyright: ignore
        if type(t_xs) is ad.Zero:
            t_xs = jnp.zeros(t_xs.aval.shape, t_xs.aval.dtype)  # pyright: ignore
        t_out = _maybe_set(
            pred, t_xs, t_x, i, kwargs=dict(kwargs), makes_false_steps=makes_false_steps
        )
    return [out], [t_out]


def _maybe_set_transpose(
    ct_out,
    pred,
    xs,
    x,
    *i_dynamic_leaves,
    i_static,
    i_treedef,
    kwargs,
    makes_false_steps,
):
    assert not ad.is_undefined_primal(pred)
    for z in i_dynamic_leaves:
        assert not ad.is_undefined_primal(z)
    i_static = _to_raw_slice(i_static)
    i = combine(i_static, jtu.tree_unflatten(i_treedef, i_dynamic_leaves))
    [ct_out] = ct_out
    if ad.is_undefined_primal(xs):
        # Not updating a zero! `_Buffer`s are write-once to each location, so when
        # transposed they are read-once from each location, meaning if we read from `i`
        # now then we will never do so again, so we can skip placing a zero there.
        #
        # Besides just being a nice efficiency gain: this is actually important for
        # working around b/288798733, in which having extra dynamic_update_slices causes
        # operations to happen out-of-place.
        ct_xs = ct_out
    else:
        ct_xs = None
    if ad.is_undefined_primal(x):
        if type(ct_out) is ad.Zero:
            ct_x = None
        else:
            ct_x = ct_out.at[i].get(**dict(kwargs))
            ct_x = _select_if_vmap(pred, ct_x, jnp.zeros_like(ct_x), makes_false_steps)
    else:
        ct_x = None
    return [None, ct_xs, ct_x] + [None] * len(i_dynamic_leaves)


maybe_set_p = create_vprim(
    "maybe_set",
    _maybe_set_impl,
    _maybe_set_abstract,
    _maybe_set_jvp,
    _maybe_set_transpose,
)


class _HashableSlice(Module):
    start: Any
    stop: Any
    step: Any


def _is_hashable_slice(x):
    return isinstance(x, _HashableSlice)


def _to_hashable_slice_impl(x):
    if isinstance(x, slice):
        return _HashableSlice(x.start, x.stop, x.step)
    else:
        return x


def _to_raw_slice_impl(x):
    if _is_hashable_slice(x):
        return slice(x.start, x.stop, x.step)
    else:
        return x


def _to_hashable_slice(tree):
    return jtu.tree_map(_to_hashable_slice_impl, tree)


def _to_raw_slice(tree):
    return jtu.tree_map(_to_raw_slice_impl, tree, is_leaf=_is_hashable_slice)


# This is a carefully optimised routine, that relies on the special behaviour exhibited
# by `_Buffer`.
# The main one is fact that `_Buffer`s are write-once to each location, so they
# are read-once to each location when transposed. This means we can skip placing zeros
# in the cotangent of `xs`. This reduces the number of in-place updates on the backward
# pass from 2 to 0. This ends up being really important for efficiency, due to
# b/288798733.
# Second, the fact that unbatched `pred` are necessarily always True (due to being
# used inside a while loop) means that we use `_select_if_vmap` over simply
# `lax.select`.
def _maybe_set(pred, xs, x, i, *, kwargs, makes_false_steps):
    """As `lax.select(pred, xs.at[i].set(x, **kwargs), xs)`, under the assumption that
    every location `i` is written to at most once. (So that we can have a more efficient
    transpose rule. Also assumes that non-vmap'd `pred` is always `True`.)
    """
    if jnp.shape(pred) != () or jnp.result_type(pred) != jnp.bool_:
        raise ValueError("predicate must be a boolean scalar.")
    dtype = jnp.result_type(x, xs)
    if dtype != jnp.result_type(xs):
        raise ValueError(
            "When doing `buffer.at[i].set(value)`, then `value` must have a dtype that "
            "can be promoted to the same dtype as `buffer`."
        )
    x = fixed_asarray(x).astype(dtype)
    x = jnp.broadcast_to(x, jax.eval_shape(lambda: xs[i]).shape)
    i_dynamic, i_static = partition(i, is_array)
    i_static = _to_hashable_slice(i_static)
    i_dynamic_leaves, i_treedef = jtu.tree_flatten(i_dynamic)
    [out] = maybe_set_p.bind(
        pred,
        xs,
        x,
        *i_dynamic_leaves,
        i_static=i_static,
        i_treedef=i_treedef,
        kwargs=tuple(kwargs.items()),
        makes_false_steps=makes_false_steps,
    )
    return out


if TYPE_CHECKING:
    from typing import Annotated, TypeAlias, TypeVar

    _T = TypeVar("_T")
    MaybeBuffer: TypeAlias = Annotated[_T, "MaybeBuffer"]
else:

    class _MetaBufferItem(type):
        def __instancecheck__(cls, instance):
            annotation = cls.annotation
            while type(instance) is _Buffer:
                instance = instance._array
            return isinstance(instance, annotation)

    class MaybeBuffer:
        def __class_getitem__(cls, item):
            class _BufferItem(metaclass=_MetaBufferItem):
                annotation = item

            return _BufferItem


class _Buffer(Module):
    # annotation removed because beartype can't handle the forward reference.
    _array: Any  # Union[Shaped[Array, "..."], _Buffer]
    _pred: Bool[Array, ""]
    _tag: object = field(static=True)
    _makes_false_steps: bool = field(static=True)

    def __getitem__(self, item):
        return self._array[item]

    def _op(self, pred, item, x, op, kwargs, makes_false_steps):
        pred = pred & self._pred
        if isinstance(self._array, _Buffer):
            array = self._array._op(pred, item, x, op, kwargs, makes_false_steps)
        else:
            array = op(
                pred,
                self._array,
                x,
                item,
                kwargs=kwargs,
                makes_false_steps=makes_false_steps,
            )
        return _Buffer(array, self._pred, self._tag, self._makes_false_steps)

    @property
    def at(self):
        return _BufferAt(self, self._makes_false_steps)

    @property
    def shape(self):
        return self._array.shape

    @property
    def dtype(self):
        return self._array.dtype

    @property
    def size(self):
        return self._array.size


class _BufferAt(Module):
    _buffer: _Buffer
    _makes_false_steps: bool = field(static=True)

    def __getitem__(self, item):
        return _BufferItem(self._buffer, item, self._makes_false_steps)


class _BufferItem(Module):
    _buffer: _Buffer
    _item: Any
    _makes_false_steps: bool = field(static=True)

    def set(self, x, *, pred=True, **kwargs):
        if pred is True:
            makes_false_steps = self._makes_false_steps
        else:
            makes_false_steps = True
        return self._buffer._op(
            pred, self._item, x, _maybe_set, kwargs, makes_false_steps
        )


def buffer_at_set(buffer: Array | _Buffer, item, x, *, pred=True, **kwargs):
    """As `buffer.at[...].set(...)`, and supports the `pred` argument even if it is an
    array.

    This is primarily useful when calling a buffer-using cond or body function outside
    of a while loop, for any reason.
    """
    if isinstance(buffer, _Buffer):
        return buffer.at[item].set(x, pred=pred, **kwargs)
    else:
        if pred is not True:
            x = jnp.where(pred, x, buffer.at[item].get(**kwargs))
        return buffer.at[item].set(x)


def _is_buffer(x):
    return isinstance(x, _Buffer)


def _unwrap_buffers(x):
    while _is_buffer(x):
        x = x._array
    return x


# Work around JAX issue #15676.
# This issue arises with both JVP tracing and make_jaxpr tracing. The former can be
# handled with a custom_jvp, but the latter cannot. So we need to just call `jnp.array`
# instead.
fixed_asarray = jnp.array


def common_rewrite(cond_fun, body_fun, init_val, max_steps, buffers, makes_false_steps):
    """Handles:

    - Efficient in-place updates;
    - Efficient/correct vmap;
    - max_steps.

    The efficient in-place updates are done using buffers. These mean that we get to
    interchange `select` (from vmap) and `scatter` (from the in-place update) so that
    they happen in the efficient order. Specifically for `checkpointed_while_loop`, it
    also means we don't need to save our buffers in the list of checkpoints.

    The efficient vmap is done by always having an unvmap'd return value from
    `cond_fun`: essentially we reduce from the vmap'd case to the no-vmap case here.
    This ensures that (a) `bounded_while_loop` exhibits early-exit, as our `lax.cond`s
    don't get turned into `lax.select`s, and (b) `checkpointed_while_loop` has the
    correct logic on the backward pass. (As all checkpoints need to be in lockstep
    across the batch.)
    """

    if buffers is None:
        new_buffers = lambda is_leaf: lambda val: []
    else:

        def new_buffers(is_leaf):
            def new_buffers2(val):
                _, _, _, val = val  # ignore step and pred
                bufs = buffers(val)
                # Ignore the ._pred attribute of nested buffers.
                # This is kind of a hack: we're special-casing support for nested
                # buffers so that end users don't have to do the ._array lookup
                # themselves.
                tree = jtu.tree_map(_unwrap_buffers, bufs, is_leaf=_is_buffer)
                return jtu.tree_leaves(tree, is_leaf=is_leaf)

            return new_buffers2

    def _wrap_buffers(val, pred, tag):
        def wrap_buffer(leaf):
            if not is_array(leaf):
                raise ValueError("Only arrays can be treated as buffers.")
            return _Buffer(leaf, pred, tag, makes_false_steps)

        _, _, _, buffer_val = tree_at(
            new_buffers(None), (None, None, None, val), replace_fn=wrap_buffer
        )
        return buffer_val

    def new_cond_fun(val):
        step, pred, prev_pred, val = val
        del pred
        # Note that this is actually recomputing `pred`!
        # For some reason this is actually a minor performance optimisation.
        # See https://github.com/patrick-kidger/diffrax/issues/274
        buffer_val = _wrap_buffers(val, prev_pred, None)
        out = unvmap_any(cond_fun(buffer_val))
        if max_steps is not None:
            if type(max_steps) is not int:
                raise ValueError("`max_steps` must be a Python integer")
            out = out & (step < max_steps)
        # Need to allow being constant across the batch. This seems to arise in some
        # edge case when using `bounded_while_loop`, in which:
        # - `lax.scan` saves a copy of its state (in particular `step`) for use in the
        #    backward pass;
        # - for some reason the `jax.checkpoint` causes such state to pick up a spurious
        #   batch tracer.
        # See:
        # https://github.com/patrick-kidger/optimistix/issues/48#issuecomment-2009221739
        return nonbatchable(
            out, name="`equinox.internal.while_loop`", allow_constant_across_batch=True
        )

    def new_body_fun(val):
        tag = object()

        def is_our_buffer(node):
            return isinstance(node, _Buffer) and node._tag is tag

        def unwrap_and_select(leaf, leaf2):
            if is_our_buffer(leaf):
                assert is_our_buffer(leaf2)
                # sanity check that this is the lowest buffer, i.e. when nesting
                # multiple checkpointed_while_loops.
                assert is_array(leaf._array)
                assert is_array(leaf2._array)
                return leaf2._array
            else:
                return _select_if_vmap(pred, leaf2, leaf, makes_false_steps)

        step, pred, _, val = val
        buffer_val = _wrap_buffers(val, pred, tag)
        buffer_val2 = body_fun(buffer_val)
        # Needed to work with `disable_jit`, as then we lose the automatic
        # ArrayLike->Array cast provided by JAX's while loops.
        # The input `val` is already cast to Array below, so this matches that.
        buffer_val2 = jtu.tree_map(fixed_asarray, buffer_val2)
        # Strip `.named_shape`; c.f. Diffrax issue #246
        struct = jax.eval_shape(lambda: buffer_val)
        struct2 = jax.eval_shape(lambda: buffer_val2)
        struct = jtu.tree_map(lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), struct)
        struct2 = jtu.tree_map(
            lambda x: jax.ShapeDtypeStruct(x.shape, x.dtype), struct2
        )
        if not tree_equal(struct, struct2):
            string = tree_pformat(struct, struct_as_array=True)
            string2 = tree_pformat(struct2, struct_as_array=True)
            out = []
            for line, line2 in it.zip_longest(
                string.split("\n"), string2.split("\n"), fillvalue=""
            ):
                if line == line2:
                    out.append("  " + line)
                else:
                    out.append("- " + line)
                    out.append("+ " + line2)
            out = "\n".join(out)
            raise ValueError(
                "`body_fun` must have the same input and output structure. Difference "
                "is:\n" + out
            )
        val2 = jtu.tree_map(
            unwrap_and_select, buffer_val, buffer_val2, is_leaf=is_our_buffer
        )
        step2 = step + 1
        pred2 = pred & cond_fun(buffer_val2)
        return step2, pred2, pred, val2

    init_val = jtu.tree_map(fixed_asarray, init_val)
    init_buffer_val = _wrap_buffers(init_val, jnp.array(True), None)
    init_pred = jnp.array(cond_fun(init_buffer_val))
    new_init_val = (jnp.array(0), init_pred, jnp.array(True), init_val)

    return new_cond_fun, new_body_fun, new_init_val, new_buffers



import types
from collections.abc import Callable
from typing import Any, Generic, TYPE_CHECKING, TypeVar
from typing_extensions import ParamSpec

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import wadler_lindig as wl
from jaxtyping import PyTree

from .._module import field, Module
from .._pretty_print import tree_pformat
from .._tree import tree_at, tree_equal


_Value = TypeVar("_Value")
_P = ParamSpec("_P")
_T = TypeVar("_T")


class _Sentinel(Module):
    """A module for sentinels that can be passed dynamically."""

    pass


# Used as a sentinel in two ways: keeping track of updated `State`s, and keeping track
# of deleted initial states.
_sentinel = _Sentinel()


class StateIndex(Module, Generic[_Value]):
    """This wraps together (a) a unique dictionary key used for looking up a stateful
    value, and (b) how that stateful value should be initialised.

    !!! Example

        ```python
        class MyStatefulLayer(eqx.Module):
            index: eqx.nn.StateIndex

            def __init__(self):
                init_state = jnp.array(0)
                self.index = eqx.nn.StateIndex(init_state)

            def __call__(self, x: Array, state: eqx.nn.State) -> tuple[Array, eqx.nn.State]:
                current_state = state.get(self.index)
                new_x = x + current_state
                new_state = state.set(self.index, current_state + 1)
                return new_x, new_state
        ```

    See also e.g. the source code of built-in stateful layers like
    [`equinox.nn.BatchNorm`][] for further reference.
    """  # noqa: E501

    # Starts off as object when initialised; later replaced with a stringified
    # jax.tree_util.KeyPath inside `make_with_state`.
    marker: object | str = field(static=True)
    init: _Value | _Sentinel

    def __init__(self, init: _Value):
        """**Arguments:**

        - `init`: The initial value for the state.
        """
        if isinstance(init, types.FunctionType):
            # Technically a function is valid here, since we could allow any pytree.
            # In practice that's weird / kind of useless, so better to explicitly raise
            # the deprecation error.
            raise ValueError(
                "As of Equinox v0.11.0, `eqx.nn.StateIndex` now accepts the value "
                "of the initial state directly. (Not a function that creates the "
                "initial state.)"
            )
        self.marker = object()
        self.init = init


def _is_index(x: Any) -> bool:
    return isinstance(x, StateIndex)


_state_error = """
Attempted to use old state. Probably you have done something like:
```
x, state2 = layer1(x, state1)
x, state3 = layer1(x, state1)  # bug! Re-used state1 instead of using state2.
```

If you have done this intentionally, because you want to use an old state, then you can
avoid this error by making a clone of the state:
```
leaves, treedef = jax.tree_util.tree_flatten(state)
state_clone = jax.tree_util.tree_unflatten(treedef, leaves)
```
""".strip()


# Basically just a dictionary which (a) works only with StateIndex-s, and which (b)
# works around a JAX bug that prevents flattening dicts with `object()` keys, and which
# (c) does error-checking that you're using the most up-to-date version of it.
@jtu.register_pytree_node_class
class State:
    """Stores the state of a model. For example, the running statistics of all
    [`equinox.nn.BatchNorm`][] layers in the model.

    This is essentially a dictionary mapping from [`equinox.nn.StateIndex`][]s to
    PyTrees of arrays.

    This class should be initialised via [`equinox.nn.make_with_state`][].
    """

    def __init__(self, model: PyTree):
        """**Arguments:**

        - `model`: any PyTree. All stateful layers (e.g. [`equinox.nn.BatchNorm`][])
            will have their initial state stored inside the `State` object.
        """
        # Note that de/serialisation depends on the ordered-ness of this dictionary,
        # between serialisation and deserialisation.
        state = {}
        leaves = jtu.tree_leaves(model, is_leaf=_is_index)
        for leaf in leaves:
            if _is_index(leaf):
                if isinstance(leaf.init, _Sentinel):
                    raise ValueError(
                        "Do not call `eqx.nn.State(model)` directly. You should call "
                        "`eqx.nn.make_with_state(ModelClass)(...args...)` instead."
                    )
                state[leaf.marker] = jtu.tree_map(jnp.asarray, leaf.init)
        self._state: _Sentinel | dict[object | int, Any] = state

    def get(self, item: StateIndex[_Value]) -> _Value:
        """Given an [`equinox.nn.StateIndex`][], returns the value of its state.

        **Arguments:**

        - `item`: an [`equinox.nn.StateIndex`][].

        **Returns:**

        The current state associated with that index.
        """
        if isinstance(self._state, _Sentinel):
            raise ValueError(_state_error)
        if type(item) is not StateIndex:
            raise ValueError("Can only use `eqx.nn.StateIndex`s as state keys.")
        return self._state[item.marker]

    def set(self, item: StateIndex[_Value], value: _Value) -> "State":
        """Sets a new value for an [`equinox.nn.StateIndex`][], **and returns the
        updated state**.

        **Arguments:**

        - `item`: an [`equinox.nn.StateIndex`][].
        - `value`: the new value associated with that index.

        **Returns:**

        A new [`equinox.nn.State`][] object, with the update.

        As a safety guard against accidentally writing `state.set(item, value)` without
        assigning it to a new value, then the old object (`self`) will become invalid.
        """
        if isinstance(self._state, _Sentinel):
            raise ValueError(_state_error)
        if type(item) is not StateIndex:
            raise ValueError("Can only use `eqx.nn.StateIndex`s as state keys.")
        old_value = self._state[item.marker]
        value = jtu.tree_map(jnp.asarray, value)
        old_struct = jax.eval_shape(lambda: old_value)
        new_struct = jax.eval_shape(lambda: value)
        if tree_equal(old_struct, new_struct) is not True:
            old_repr = tree_pformat(old_struct, struct_as_array=True)
            new_repr = tree_pformat(new_struct, struct_as_array=True)
            raise ValueError(
                "Old and new values have different structures/shapes/dtypes. The old "
                f"value is {old_repr} and the new value is {new_repr}."
            )
        state = self._state.copy()  # pyright: ignore
        state[item.marker] = value
        new_self = object.__new__(State)
        new_self._state = state
        self._state = _sentinel
        return new_self

    def substate(self, pytree: PyTree) -> "State":
        """Creates a smaller `State` object, that tracks only the states of some smaller
        part of the overall model.

        **Arguments:**

        - `pytree`: any PyTree. It will be iterated over to check for
            [`equinox.nn.StateIndex`]s.

        **Returns:**

        A new [`equinox.nn.State`][] object, which tracks only some of the overall
        states.
        """
        if isinstance(self._state, _Sentinel):
            raise ValueError(_state_error)
        leaves = jtu.tree_leaves(pytree, is_leaf=_is_index)
        markers = [x.marker for x in leaves if _is_index(x)]
        substate = {k: self._state[k] for k in markers}  # pyright: ignore
        subself = object.__new__(State)
        subself._state = substate
        return subself

    def update(self, substate: "State") -> "State":
        """Takes a smaller `State` object (typically produces via `.substate`), and
        updates states by using all of its values.

        **Arguments:**

        - `substate`: a `State` object whose keys are a subset of the keys of `self`.

        **Returns:**

        A new [`equinox.nn.State`][] object, containing all of the updated values.

        As a safety guard against accidentally writing `state.set(item, value)` without
        assigning it to a new value, then the old object (`self`) will become invalid.
        """
        if isinstance(self._state, _Sentinel):
            raise ValueError(_state_error)
        if type(substate) is not State:
            raise ValueError("Can only use `eqx.nn.State`s in `update`.")
        state = self._state.copy()  # pyright: ignore
        for key, value in substate._state.items():  # pyright: ignore
            if key not in state:
                raise ValueError(
                    "Cannot call `state1.update(state2)` unless `state2` is a substate "
                    "of `state1`."
                )
            state[key] = value
        new_self = object.__new__(State)
        new_self._state = state
        self._state = _sentinel
        return new_self

    def __repr__(self):
        return tree_pformat(self)

    def __pdoc__(self, **kwargs):
        if isinstance(self._state, _Sentinel):
            return wl.TextDoc("State(~old~)")
        else:
            docs = wl.named_objs(self._state.items(), **kwargs)  # pyright: ignore
            return wl.bracketed(
                begin=wl.TextDoc("State("),
                docs=docs,
                sep=wl.comma,
                end=wl.TextDoc(")"),
                indent=kwargs["indent"],
            )

    def tree_flatten(self):
        if isinstance(self._state, _Sentinel):
            raise ValueError(_state_error)
        keys = tuple(self._state.keys())  # pyright: ignore
        values = tuple(self._state[k] for k in keys)  # pyright: ignore
        return values, keys

    @classmethod
    def tree_unflatten(cls, keys, values):
        self = object.__new__(cls)
        state = {}
        assert len(keys) == len(values)
        for key, value in zip(keys, values):
            state[key] = value
        self._state = state
        return self


def _delete_init_state(x):
    if _is_index(x):
        return tree_at(lambda y: y.init, x, _sentinel)
    else:
        return x


def delete_init_state(model: PyTree) -> PyTree:
    """For memory efficiency, this deletes the initial state stored within a model.

    Every stateful layer keeps a copy of the initial value of its state. This is then
    collected by [`equinox.nn.State`][], when it is called on the model. However, this
    means that the model must keep a copy of the initial state around, in case
    `eqx.nn.State` is called on it again. This extra copy consumes extra memory.

    But in practice, it is quite common to only need to initialise the state once. In
    this case, we can use this function to delete this extra copy, and in doing so save
    some memory.

    !!! Example

        Here is the typical pattern in which this is used:
        ```python
        model_and_state = eqx.nn.BatchNorm(...)
        state = eqx.nn.State(model_and_state)
        model = eqx.nn.delete_init_state(model)
        del model_and_state  # ensure this goes out of scope and is garbage collected
        ```
        Indeed the above is precisely what [`equinox.nn.make_with_state`][] does.

    **Arguments:**

    - `model`: any PyTree.

    **Returns:**

    A copy of `model`, with all the initial states stripped out. (As in the examples
    above, you should then dispose of the original `model` object.)
    """
    return jtu.tree_map(_delete_init_state, model, is_leaf=_is_index)


def make_with_state(make_model: Callable[_P, _T]) -> Callable[_P, tuple[_T, State]]:
    """This function is the most common API for working with stateful models. This
    initialises both the parameters and the state of a stateful model.

    `eqx.nn.make_with_state(Model)(*args, **kwargs)` simply calls
    `model_with_state = Model(*args, **kwargs)`, and then partitions the resulting
    PyTree into two pieces: the parameters, and the state.

    **Arguments:**

    - `make_model`: some callable returning a PyTree.

    **Returns:**

    A callable, which when evaluated returns a 2-tuple of `(model, state)`, where
    `model` is the result of `make_model(*args, **kwargs)` but with all of the initial
    states stripped out, and `state` is an [`equinox.nn.State`][] object encapsulating
    the initial states.

    !!! Example

        See [the stateful example](../../examples/stateful.ipynb) for a runnable
        example.

        ```python
        class Model(eqx.Module):
            def __init__(self, foo, bar):
                ...

            ...

        model, state = eqx.nn.make_with_state(Model)(foo=3, bar=4)
        ```
    """

    # _P.{args, kwargs} not supported by beartype
    if TYPE_CHECKING:

        def make_with_state_impl(
            *args: _P.args, **kwargs: _P.kwargs
        ) -> tuple[_T, State]: ...

    else:

        def make_with_state_impl(*args, **kwargs) -> tuple[_T, State]:
            model = make_model(*args, **kwargs)

            # Replace all markers with stringified key paths. This is needed to ensure
            # that two calls to `make_with_state` produce compatible models and states.
            key_leaves, treedef = jtu.tree_flatten_with_path(model, is_leaf=_is_index)
            new_leaves = []
            for key, leaf in key_leaves:
                if _is_index(leaf):
                    leaf = StateIndex(leaf.init)
                    path_str = model.__class__.__name__ + jtu.keystr(key)
                    object.__setattr__(leaf, "marker", path_str)
                new_leaves.append(leaf)
            model = jtu.tree_unflatten(treedef, new_leaves)

            state = State(model)
            model = delete_init_state(model)
            return model, state

    return make_with_state_impl



import functools as ft
import inspect
import weakref
from collections.abc import Callable
from typing import overload, TypeVar
from typing_extensions import ParamSpec

from .._custom_types import sentinel
from .._eval_shape import filter_eval_shape
from .._module import field, Module, module_update_wrapper
from .._pretty_print import tree_pformat
from .._tree import tree_equal


_T = TypeVar("_T")
_P = ParamSpec("_P")


# If we wanted we could actually store this directly on the `_AssertMaxTraces` object.
# Replace `tag` with a list of a single integer, `[0]`, and propagate the list as static
# metadata so that even after flattening and unflattening they share a single counter.
#
# In practice we do it this way to make it abundantly clear that this counter is shared
# across cloned (flattened+unflattened) instances.
_traces = weakref.WeakKeyDictionary()
_seen_args = weakref.WeakKeyDictionary()


class _Weakrefable:
    __slots__ = ("__weakref__",)


class _AssertMaxTraces(Module):
    fn: Callable
    max_traces: int | None = field(static=True)
    tag: _Weakrefable = field(static=True)

    def __init__(self, fn, max_traces):
        self.fn = fn
        self.max_traces = max_traces
        self.tag = _Weakrefable()
        _traces[self.tag] = 0
        _seen_args[self.tag] = {}

    @property
    def __wrapped__(self):
        return self.fn

    def __call__(self, *args, **kwargs):
        __tracebackhide__ = True
        num_traces = _traces[self.tag] = _traces[self.tag] + 1
        arguments = inspect.signature(self.fn).bind(*args, **kwargs).arguments
        if self.max_traces is not None and num_traces > self.max_traces:
            for name, value in arguments.items():
                struct = filter_eval_shape(lambda: value)
                seen_structs = _seen_args[self.tag].get(name, [])
                for seen_struct in seen_structs:
                    if tree_equal(struct, seen_struct) is True:
                        break
                else:
                    struct_str = tree_pformat(struct, struct_as_array=True)
                    seen_struct_str = "\n".join(
                        tree_pformat(seen_struct, struct_as_array=True)
                        for seen_struct in seen_structs
                    )
                    raise RuntimeError(
                        f"{self.fn} can only be traced {self.max_traces} times. "
                        f"However, it is has now been traced {num_traces} times. It "
                        f"appears that the '{name}' argument is responsible: it "
                        f"currently has the value:\n{struct_str}\nPrevious values are:"
                        f"\n{seen_struct_str}"
                    )
            raise RuntimeError(
                f"{self.fn} can only be traced {self.max_traces} times. However, "
                f"it is has now been traced {num_traces} times. Could not determine "
                "argument was responsible for re-tracing."
            )
        for name, value in arguments.items():
            struct = filter_eval_shape(lambda: value)
            if name not in _seen_args[self.tag]:
                _seen_args[self.tag][name] = []
            _seen_args[self.tag][name].append(struct)
        return self.fn(*args, **kwargs)


@overload
def assert_max_traces(
    fn: Callable[_P, _T], *, max_traces: int | None
) -> Callable[_P, _T]: ...


@overload
def assert_max_traces(
    *,
    max_traces: int | None,
) -> Callable[[Callable[_P, _T]], Callable[_P, _T]]: ...


def assert_max_traces(fn: Callable = sentinel, *, max_traces: int | None):
    """Asserts that the wrapped callable is not called more than `max_traces` times.

    The typical use-case for this is to check that a JIT-compiled function is not
    compiled more than `max_traces` times. (I.e. this function can be used to guard
    against bugs.) In this case it should be placed within the JIT wrapper.

    !!! Example

        ```python
        @eqx.filter_jit
        @eqx.debug.assert_max_traces(max_traces=1)
        def f(x, y):
            return x + y
        ```

    **Arguments:**

    - `fn`: The callable to wrap.
    - `max_traces`: keyword only argument. The maximum number of calls that are
        allowed. Can be `None` to allow arbitrarily many calls; in this case the number
        of calls can still can be found via [`equinox.debug.get_num_traces`][].

    **Returns:**

    A wrapped version of `fn` that tracks the number of times it is called. This will
    raise a `RuntimeError` if is called more than `max_traces` many times.

    !!! info

        See also
        [`chex.assert_max_traces`](https://github.com/google-deepmind/chex#jax-tracing-assertions)
        which provides similar functionality.

        The differences are that (a) Chex's implementation is a bit stricter, as the
        following will raise:
        ```python
        import chex
        import jax

        def f(x):
            pass

        f2 = jax.jit(chex.assert_max_traces(f, 1))
        f3 = jax.jit(chex.assert_max_traces(f, 1))

        f2(1)
        f3(1)  # will raise, despite the fact that f2 and f3 are different.
        ```
        and (b) Equinox's implementation supports non-function callables.

        You may prefer the Chex version if you prefer the stricter raising behaviour of
        the above code.
    """
    if fn is sentinel:
        return ft.partial(assert_max_traces, max_traces=max_traces)
    return module_update_wrapper(_AssertMaxTraces(fn, max_traces))


def get_num_traces(fn) -> int:
    """Given a function wrapped in [`equinox.debug.assert_max_traces`][], return the
    number of times which it has been traced so far.
    """
    num_traces = _get_num_traces(fn)
    if num_traces is None:
        raise ValueError(
            f"{fn} does not appear to wrapped with `eqx.debug.assert_max_traces`"
        )
    return num_traces


def _get_num_traces(fn) -> int | None:
    if isinstance(fn, _AssertMaxTraces):
        return _traces[fn.tag]
    elif hasattr(fn, "__wrapped__"):
        return _get_num_traces(fn.__wrapped__)
    else:
        return None



from typing import Any

import jax
import jax._src.traceback_util as traceback_util
import jax.core
import jax.numpy as jnp
import jax.tree_util as jtu

from .._ad import filter_custom_vjp
from .._errors import error_if
from .._filters import filter, is_array_like
from .._module import Module
from .._pretty_print import tree_pformat


traceback_util.register_exclusion(__file__)


def backward_nan(x, name=None, terminate=True):
    """Debug NaNs that only occur on the backward pass.

    **Arguments:**

    - `x`: a variable to intercept.
    - `name`: an optional name to appear in printed debug statements.
    - `terminate`: whether to halt the computation if a NaN cotangent is found. If
        `True` then an error will be raised via [`equinox.error_if`][]. (So you can
        also arrange for a breakpoint to trigger by setting `EQX_ON_ERROR`
        appropriately.)

    **Returns:**

    The `x` argument is returned unchanged.

    As a side-effect, both the primal and the cotangent for `x` will be printed out
    during the backward pass.
    """
    return _backward_nan(x, name, terminate)


@filter_custom_vjp
def _backward_nan(x, name, terminate):
    return x


@_backward_nan.def_fwd
def _backward_nan_fwd(perturbed, x, name, terminate):
    del perturbed
    return backward_nan(x, name, terminate), None


class _LongRepr(Module):
    obj: Any

    def __repr__(self):
        return tree_pformat(self.obj, short_arrays=False)


@_backward_nan.def_bwd
def _backward_nan_bwd(residuals, grad_x, perturbed, x, name, terminate):
    del residuals, perturbed
    msg = "   primals={x}\ncotangents={grad_x}"
    if name is not None:
        msg = f"{name}:\n" + msg
    jax.debug.print(msg, x=_LongRepr(x), grad_x=_LongRepr(grad_x), ordered=True)
    if terminate:
        nans = [
            jnp.isnan(a).any() for a in jtu.tree_leaves(filter(grad_x, is_array_like))
        ]
        grad_x = error_if(grad_x, jnp.any(jnp.stack(nans)), "Encountered NaN")
    return grad_x



from collections.abc import Callable, Sequence
from typing import Any, TYPE_CHECKING

import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np
from jaxtyping import Array, ArrayLike, Bool, Float, PyTree, PyTreeDef

from ._custom_types import sentinel
from ._doc_utils import doc_repr
from ._filters import is_array


if TYPE_CHECKING:
    _Node = Any
else:
    _Node = doc_repr(Any, "Node")


class _LeafWrapper:
    def __init__(self, value: Any):
        self.value = value


def _remove_leaf_wrapper(x: _LeafWrapper) -> Any:
    if not isinstance(x, _LeafWrapper):
        raise TypeError(f"Operation undefined, {x} is not a leaf of the pytree.")
    return x.value


class _CountedIdDict:
    def __init__(self, keys, values):
        assert len(keys) == len(values)
        self._dict = {id(k): v for k, v in zip(keys, values)}
        self._count = {id(k): 0 for k in keys}

    def __contains__(self, item):
        return id(item) in self._dict

    def __getitem__(self, item):
        self._count[id(item)] += 1
        return self._dict[id(item)]

    def get(self, item, default):
        try:
            return self[item]
        except KeyError:
            return default

    def count(self, item):
        return self._count[id(item)]


class _DistinctTuple(tuple):
    pass


def tree_at(
    where: Callable[[PyTree], _Node | Sequence[_Node]],
    pytree: PyTree,
    replace: Any | Sequence[Any] = sentinel,
    replace_fn: Callable[[_Node], Any] = sentinel,
    is_leaf: Callable[[Any], bool] | None = None,
):
    """Modifies a leaf or subtree of a PyTree. (A bit like using `.at[].set()` on a JAX
    array.)

    The modified PyTree is returned and the original input is left unchanged. Make sure
    to use the return value from this function!

    **Arguments:**

    - `where`: A callable `PyTree -> Node` or `PyTree -> tuple[Node, ...]`. It should
        consume a PyTree with the same structure as `pytree`, and return the node or
        nodes that should be replaced. For example
        `where = lambda mlp: mlp.layers[-1].linear.weight`.
    - `pytree`: The PyTree to modify.
    - `replace`: Either a single element, or a sequence of the same length as returned
        by `where`. This specifies the replacements to make at the locations specified
        by `where`. Mutually exclusive with `replace_fn`.
    - `replace_fn`: A function `Node -> Any`. It will be called on every node specified
        by `where`. The return value from `replace_fn` will be used in its place.
        Mutually exclusive with `replace`.
    - `is_leaf`: As `jtu.tree_flatten`. For example pass `is_leaf=lambda x: x is None`
        to be able to replace `None` values using `tree_at`.

    Note that `where` should not depend on the type of any of the leaves of the
    pytree, e.g. given `pytree = [1, 2, object(), 3]`, then
    `where = lambda x: tuple(xi for xi in x if type(xi) is int)` is not allowed. If you
    really need this behaviour then this example could instead be expressed as
    `where = lambda x: tuple(xi for xi, yi in zip(x, pytree) if type(yi) is int)`.

    **Returns:**

    A new PyTree with the same structure as the input PyTree and the appropriate modifications.

    (If donating JAX arrays on JIT boundaries, then note that this function does not make a copy of the JAX arrays.)

    !!! Example

        ```python
        # Here is a pytree
        tree = [1, [2, {"a": 3, "b": 4}]]
        new_leaf = 5
        get_leaf = lambda t: t[1][1]["a"]
        new_tree = eqx.tree_at(get_leaf, tree, 5)
        # new_tree is [1, [2, {"a": 5, "b": 4}]]
        # The original tree is unchanged.
        ```

    !!! Example

        This is useful for performing model surgery. For example:
        ```python
        mlp = eqx.nn.MLP(...)
        new_linear = eqx.nn.Linear(...)
        get_last_layer = lambda m: m.layers[-1]
        new_mlp = eqx.tree_at(get_last_layer, mlp, new_linear)
        ```
        See also the [Tricks](../tricks.md) page.

    !!! Info

        Constructing analogous PyTrees, with the same structure but different leaves, is
        very common in JAX: for example when constructing the `in_axes` argument to
        `jax.vmap`.

        To support this use-case, the returned PyTree is constructed without calling
        `__init__`, `__post_init__`, or
        [`__check_init__`](./module/advanced_fields.md#checking-invariants). This allows
        for modifying leaves to be anything, regardless of the use of any custom
        constructor or custom checks in the original PyTree.
    """  # noqa: E501

    # We need to specify a particular node in a PyTree.
    # This is surprisingly difficult to do! As far as I can see, pretty much the only
    # way of doing this is to specify e.g. `x.foo[0].bar` via `is`, and then pulling
    # a few tricks to try and ensure that the same object doesn't appear multiple
    # times in the same PyTree.
    #
    # So this first `tree_map` serves a dual purpose.
    # 1) Makes a copy of the composite nodes in the PyTree, to avoid aliasing via
    #    e.g. `pytree=[(1,)] * 5`. This has the tuple `(1,)` appear multiple times.
    # 2) It makes each leaf be a unique Python object, as it's wrapped in
    #    `_LeafWrapper`. This is needed because Python caches a few builtin objects:
    #    `assert 0 + 1 is 1`. I think only a few leaf types are subject to this.
    # So point 1) should ensure that all composite nodes are unique Python objects,
    # and point 2) should ensure that all leaves are unique Python objects.
    # Between them, all nodes of `pytree` are handled.
    #
    # I think pretty much the only way this can fail is when using a custom node with
    # singleton-like flatten+unflatten behaviour, which is pretty edge case. And we've
    # added a check for it at the bottom of this function, just to be sure.
    #
    # Whilst we're here: we also double-check that `where` is well-formed and doesn't
    # use leaf information. (As else `node_or_nodes` will be wrong.)
    is_empty_tuple = (
        lambda x: isinstance(x, tuple) and not hasattr(x, "_fields") and x == ()
    )
    pytree = jtu.tree_map(
        lambda x: _DistinctTuple() if is_empty_tuple(x) else x,
        pytree,
        is_leaf=is_empty_tuple,
    )
    node_or_nodes_nowrapper = where(pytree)
    pytree = jtu.tree_map(_LeafWrapper, pytree, is_leaf=is_leaf)
    node_or_nodes = where(pytree)
    leaves1, structure1 = jtu.tree_flatten(node_or_nodes_nowrapper, is_leaf=is_leaf)
    leaves2, structure2 = jtu.tree_flatten(node_or_nodes)
    leaves2 = [_remove_leaf_wrapper(x) for x in leaves2]
    if (
        structure1 != structure2
        or len(leaves1) != len(leaves2)
        or any(l1 is not l2 for l1, l2 in zip(leaves1, leaves2))
    ):
        raise ValueError(
            "`where` must use just the PyTree structure of `pytree`. `where` must not "
            "depend on the leaves in `pytree`."
        )
    del node_or_nodes_nowrapper, leaves1, structure1, leaves2, structure2

    # Normalise whether we were passed a single node or a sequence of nodes.
    in_pytree = False

    def _in_pytree(x):
        nonlocal in_pytree
        if x is node_or_nodes:  # noqa: F821
            in_pytree = True
        return x  # needed for jax.tree_util.Partial, which has a dodgy constructor

    jtu.tree_map(_in_pytree, pytree, is_leaf=lambda x: x is node_or_nodes)  # noqa: F821
    if in_pytree:
        nodes = (node_or_nodes,)
        if replace is not sentinel:
            replace = (replace,)
    else:
        nodes = node_or_nodes
    del in_pytree, node_or_nodes

    # Normalise replace vs replace_fn
    if replace is sentinel:
        if replace_fn is sentinel:
            raise ValueError(
                "Precisely one of `replace` and `replace_fn` must be specified."
            )
        else:

            def _replace_fn(x):
                x = jtu.tree_map(_remove_leaf_wrapper, x)
                return replace_fn(x)

            replace_fns = [_replace_fn] * len(nodes)
    else:
        if replace_fn is sentinel:
            if len(nodes) != len(replace):
                raise ValueError(
                    "`where` must return a sequence of leaves of the same length as "
                    "`replace`."
                )
            replace_fns = [lambda _, r=r: r for r in replace]
        else:
            raise ValueError(
                "Precisely one of `replace` and `replace_fn` must be specified."
            )
    node_replace_fns = _CountedIdDict(nodes, replace_fns)

    # Actually do the replacement
    def _make_replacement(x: _Node) -> Any:
        return node_replace_fns.get(x, _remove_leaf_wrapper)(x)

    out = jtu.tree_map(
        _make_replacement, pytree, is_leaf=lambda x: x in node_replace_fns
    )

    # Check that `where` is well-formed.
    for node in nodes:
        count = node_replace_fns.count(node)
        if count == 0:
            raise ValueError(
                "`where` does not specify an element or elements of `pytree`."
            )
        elif count == 1:
            pass
        else:
            raise ValueError(
                "`where` does not uniquely identify a single element of `pytree`. This "
                "usually occurs when trying to replace a `None` value:\n"
                "\n"
                "  >>> eqx.tree_at(lambda t: t[0], (None, None, 1), True)\n"
                "\n"
                "\n"
                "for which the fix is to specify that `None`s should be treated as "
                "leaves:\n"
                "\n"
                "  >>> eqx.tree_at(lambda t: t[0], (None, None, 1), True,\n"
                "  ...             is_leaf=lambda x: x is None)"
            )
    out = jtu.tree_map(
        lambda x: () if isinstance(x, _DistinctTuple) else x,
        out,
        is_leaf=lambda x: isinstance(x, _DistinctTuple),
    )
    return out


def _array_equal(x, y, npi, rtol, atol):
    assert x.dtype == y.dtype
    if (
        isinstance(rtol, (int, float))
        and isinstance(atol, (int, float))
        and rtol == 0
        and atol == 0
    ) or not npi.issubdtype(x.dtype, npi.inexact):
        return npi.all(x == y)
    else:
        return npi.allclose(x, y, rtol=rtol, atol=atol)


def tree_equal(
    *pytrees: PyTree,
    typematch: bool = False,
    rtol: Float[ArrayLike, ""] = 0.0,
    atol: Float[ArrayLike, ""] = 0.0,
) -> bool | Bool[Array, ""]:
    """Returns `True` if all input PyTrees are equal. Every PyTree must have the same
    structure, and all leaves must be equal.

    - For JAX arrays, NumPy arrays, or NumPy scalars: they must have the same shape and
        dtype.
        - If `rtol=0` and `atol=0` (the default) then all their values must be equal.
            Otherwise they must satisfy
            `{j}np.allclose(leaf1, leaf2, rtol=rtol, atol=atol)`.
        - If `typematch=False` (the default) then JAX and NumPy arrays are considered
            equal to each other. If `typematch=True` then JAX and NumPy are not
            considered equal to each other.
    - For non-arrays, if `typematch=False` (the default) then equality is determined
        with just `leaf1 == leaf2`. If `typematch=True` then
        `type(leaf1) == type(leaf2)` is also required.

    If used under JIT, and any JAX arrays are present, then this may return a tracer.
    Use the idiom `result = tree_equal(...) is True` if you'd like to assert that the
    result is statically true without dependence on the value of any traced arrays.

    **Arguments:**

    - `*pytrees`: Any number of PyTrees each with any structure.
    - `typematch:` Whether to additionally require that corresponding leaves should have
        the same `type(...)` as each other.
    - `rtol`: Used to determine the `rtol` of `jnp.allclose`/`np.allclose` when
        comparing inexact (floating or complex) arrays. Defaults to zero, i.e. requires
        exact equality.
    - `atol`: As `rtol`.

    **Returns:**

    A boolean, or bool-typed tracer.
    """
    flat, treedef = jtu.tree_flatten(pytrees[0])
    traced_out = True
    for pytree in pytrees[1:]:
        flat_, treedef_ = jtu.tree_flatten(pytree)
        if treedef_ != treedef:
            return False
        assert len(flat) == len(flat_)
        for elem, elem_ in zip(flat, flat_):
            if typematch:
                if type(elem) != type(elem_):
                    return False
            if isinstance(elem, (np.ndarray, np.generic)) and isinstance(
                elem_, (np.ndarray, np.generic)
            ):
                if (
                    (elem.shape != elem_.shape)
                    or (elem.dtype != elem_.dtype)
                    or not _array_equal(elem, elem_, np, rtol, atol)
                ):
                    return False
            elif is_array(elem):
                if is_array(elem_):
                    if (elem.shape != elem_.shape) or (elem.dtype != elem_.dtype):
                        return False
                    traced_out = traced_out & _array_equal(elem, elem_, jnp, rtol, atol)
                else:
                    return False
            else:
                if is_array(elem_):
                    return False
                else:
                    if elem != elem_:
                        return False
    return traced_out


def tree_flatten_one_level(
    pytree: PyTree,
) -> tuple[list[PyTree], PyTreeDef]:  # pyright: ignore[reportInvalidTypeForm]
    """Returns the immediate subnodes of a PyTree node. If called on a leaf node then it
    will return just that leaf.

    **Arguments:**

    - `pytree`: the PyTree node to flatten by one level.

    **Returns:**

    As `jax.tree_util.tree_flatten`: a list of leaves and a `PyTreeDef`.

    !!! Example

        ```python
        x = {"a": 3, "b": (1, 2)}
        eqx.tree_flatten_one_level(x)
        # ([3, (1, 2)], PyTreeDef({'a': *, 'b': *}))

        y = 4
        eqx.tree_flatten_one_level(y)
        # ([4], PyTreeDef(*))
        ```
    """
    seen_pytree = False

    def is_leaf(node):
        nonlocal seen_pytree
        if node is pytree:
            if seen_pytree:
                # We expect to only see it once as the root.
                # This catches for example
                # ```python
                # x = []
                # x.append(x)
                # tree_subnodes(x)
                # ```
                # Note that it intentionally does *not* catch
                # ```python
                # x = []
                # y = []
                # x.append(y)
                # y.append(x)
                # tree_subnodes(x)
                # ```
                # as `x` is not an immediate subnode of itself.
                # If you want to check for that then use `tree_check`.
                try:
                    type_string = type(pytree).__name__
                except AttributeError:
                    type_string = "<unknown>"
                raise ValueError(
                    f"PyTree node of type `{type_string}` is immediately "
                    "self-referential; that is to say it appears within its own PyTree "
                    "structure as an immediate subnode. (For example "
                    "`x = []; x.append(x)`.) This is not allowed."
                )
            else:
                seen_pytree = True
            return False
        else:
            return True

    return jtu.tree_flatten(pytree, is_leaf=is_leaf)


def tree_check(pytree: Any) -> None:
    """Checks if the PyTree has no self-references, and if all non-leaf nodes are unique
    Python objects. (For example something like `x = [1]; y = [x, x]` would fail as `x`
    appears twice in the PyTree.)

    Having unique non-leaf nodes isn't actually a requirement that JAX imposes, but it
    will become true after passing through an operation like `jax.{jit, grad, ...}` (as
    JAX copies the PyTree without trying to preserve identity). As such some users like
    to use this function to assert that this invariant was already true prior to the
    transform, as a way to avoid surprises.

    !!! Example

        ```python
        a = 1
        eqx.tree_check([a, a])  # passes, duplicate is a leaf

        b = eqx.nn.Linear(...)
        eqx.tree_check([b, b])  # fails, duplicate is non-leaf!

        c = []  # empty list
        eqx.tree_check([c, c])  # passes, duplicate is leaf

        d = eqx.Module()
        eqx.tree_check([d, d])  # passes, duplicate is leaf

        eqx.tree_check([None, None])  # passes, duplicate is leaf

        e = [1]
        eqx.tree_check([e, e])  # fails, duplicate is non-leaf!

        eqx.tree_check([[1], [1]])  # passes, not actually a duplicate: each `[1]`
                                    # has the same structure, but they're different.

        # passes, not actually a duplicate: each Linear layer is a separate layer.
        eqx.tree_check([eqx.nn.Linear(...), eqx.nn.Linear(...)])
        ```

    **Arguments:**

    - `pytree`: the PyTree to check.

    **Returns:**

    Nothing.

    **Raises:**

    A `ValueError` if the PyTree is not well-formed.
    """
    all_nodes = {}
    _tree_check(pytree, all_nodes)


_leaf_treedef = jtu.tree_structure(0)


def _tree_check(node, all_nodes):
    subnodes, treedef = tree_flatten_one_level(node)
    # We allow duplicate leaves and empty containers, so don't raise an error with those
    if treedef != _leaf_treedef and treedef.num_leaves > 0:
        try:
            self_referential, type_string = all_nodes[id(node)]
        except KeyError:
            pass
        else:
            if self_referential:
                raise ValueError(
                    f"PyTree node of type `{type_string}` is self-referential; that is "
                    "to say it appears somewhere within its own PyTree structure. This "
                    "is not allowed."
                )
            else:
                raise ValueError(
                    f"PyTree node of type `{type_string}` appears in the PyTree "
                    "multiple times. This is almost always an error, as these nodes "
                    "will turn into two duplicate copies after "
                    "flattening/unflattening, e.g. when crossing a JIT boundary."
                )
        try:
            type_string = type(node).__name__
        except AttributeError:
            # AttributeError: in case we cannot get __name__ for some weird reason.
            type_string = "<unknown type>"
        all_nodes[id(node)] = (True, type_string)
        for subnode in subnodes:
            _tree_check(subnode, all_nodes)
        all_nodes[id(node)] = (False, type_string)



from typing import Any

import jax
import jax.lax as lax
from jaxtyping import PyTree

from ._filters import combine, is_array, partition


def filter_shard(
    x: PyTree[Any],
    device_or_shardings: jax.Device | jax.sharding.Sharding,  # pyright: ignore[reportInvalidTypeForm]
):
    """Filtered transform combining `jax.lax.with_sharding_constraint`
    and `jax.device_put`.

    Enforces sharding within a JIT'd computation (That is, how an array is
    split between multiple devices, i.e. multiple GPUs/TPUs.), or outside a
    JIT'd region moves `x` to a device.

    **Arguments:**

    - `x`: A PyTree, with potentially a mix of arrays and non-arrays on the leaves.
        They will have their shardings constrained.
    - `device_or_shardings`: Either a singular device (e.g. CPU or GPU) or PyTree of
        sharding specifications. The structure should be a prefix of `x`.

    **Returns:**

    A copy of `x` with the specified sharding constraints.

    !!! Example
        See also the [autoparallelism example](../examples/parallelism.ipynb).
    """
    if isinstance(device_or_shardings, jax.Device):
        shardings = jax.sharding.SingleDeviceSharding(device_or_shardings)
    else:
        shardings = device_or_shardings
    dynamic, static = partition(x, is_array)
    # `with_sharding_constraint` is documented in JAX for jitted
    # regions, while `device_put` is used for non-jitted regions.
    # However, it suffices to simply call `with_sharding_constraint` here!
    dynamic = lax.with_sharding_constraint(dynamic, shardings)
    return combine(dynamic, static)



import itertools as it
import math
from collections.abc import Callable, Sequence
from typing import cast, TypeVar

import jax.lax as lax
import jax.numpy as jnp
import jax.random as jrandom
import numpy as np
from jaxtyping import Array, PRNGKeyArray

from .._misc import default_floating_dtype
from .._module import field, Module
from ._misc import all_sequences, default_init, named_scope


_T = TypeVar("_T")


def _ntuple(n: int) -> Callable[[_T | Sequence[_T]], tuple[_T, ...]]:
    def parse(x: _T | Sequence[_T]) -> tuple[_T, ...]:
        if isinstance(x, Sequence):
            if len(x) == n:
                return tuple(x)
            else:
                raise ValueError(
                    f"Length of {x} (length = {len(x)}) is not equal to {n}"
                )
        else:
            return tuple(it.repeat(x, n))

    return parse


def _padding_init(
    padding: str | int | Sequence[int] | Sequence[tuple[int, int]],
    num_spatial_dims: int,
) -> str | tuple[tuple[int, int], ...]:
    if isinstance(padding, str):
        padding = padding.upper()
        if padding not in ("SAME", "SAME_LOWER", "VALID"):
            raise ValueError(
                "`padding` string must be `'SAME'`, `'SAME_LOWER'`, or `'VALID'`."
            )
    elif isinstance(padding, int):
        padding = tuple((padding, padding) for _ in range(num_spatial_dims))
    elif isinstance(padding, Sequence) and len(padding) == num_spatial_dims:
        if all_sequences(padding):
            padding = tuple(padding)  # pyright: ignore
        else:
            padding = tuple((p, p) for p in padding)
    else:
        raise ValueError(
            "`padding` must either be a string, an int, or tuple of length "
            f"{num_spatial_dims} containing ints or tuples of length 2."
        )
    return padding  # pyright: ignore


def _padding_mode_init(padding_mode: str) -> str:
    padding_mode = padding_mode.upper()
    if padding_mode not in ("ZEROS", "REFLECT", "REPLICATE", "CIRCULAR"):
        raise ValueError(
            "`padding_mode` must be `'ZEROS'`, `'REFLECT'`, `'REPLICATE'`, or "
            "`'CIRCULAR'`."
        )
    return padding_mode


class Conv(Module):
    """General N-dimensional convolution."""

    num_spatial_dims: int = field(static=True)
    weight: Array
    bias: Array | None
    in_channels: int = field(static=True)
    out_channels: int = field(static=True)
    kernel_size: tuple[int, ...] = field(static=True)
    stride: tuple[int, ...] = field(static=True)
    padding: str | tuple[tuple[int, int], ...] = field(static=True)
    dilation: tuple[int, ...] = field(static=True)
    groups: int = field(static=True)
    use_bias: bool = field(static=True)
    padding_mode: str = field(static=True)

    def __init__(
        self,
        num_spatial_dims: int,
        in_channels: int,
        out_channels: int,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = 1,
        padding: str | int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        dilation: int | Sequence[int] = 1,
        groups: int = 1,
        use_bias: bool = True,
        padding_mode: str = "ZEROS",
        dtype=None,
        *,
        key: PRNGKeyArray,
    ):
        """**Arguments:**

        - `num_spatial_dims`: The number of spatial dimensions. For example traditional
            convolutions for image processing have this set to `2`.
        - `in_channels`: The number of input channels.
        - `out_channels`: The number of output channels.
        - `kernel_size`: The size of the convolutional kernel.
        - `stride`: The stride of the convolution.
        - `padding`: The padding of the convolution.
        - `dilation`: The dilation of the convolution.
        - `groups`: The number of input channel groups. At `groups=1`,
            all input channels contribute to all output channels. Values
            higher than `1` are equivalent to running `groups` independent
            `Conv` operations side-by-side, each having access only to
            `in_channels` // `groups` input channels, and
            concatenating the results along the output channel dimension.
            `in_channels` must be divisible by `groups`.
        - `use_bias`: Whether to add on a bias after the convolution.
        - `padding_mode`: One of the following strings specifying the padding values.
            - `'ZEROS'` (default): pads with zeros, `1234 -> 00123400`.
            - `'REFLECT'`: pads with the reflection on boundary, `1234 -> 32123432`.
            - `'REPLICATE'`: pads with the replication of edge values,
                `1234 -> 11123444`.
            - `'CIRCULAR'`: pads with circular values, `1234 -> 34123412`.
        - `dtype`: The dtype to use for the weight and the bias in this layer.
            Defaults to either `jax.numpy.float32` or `jax.numpy.float64` depending
            on whether JAX is in 64-bit mode.
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)

        !!! info

            All of `kernel_size`, `stride`, `padding`, `dilation` can be either an
            integer or a sequence of integers.

            If they are an integer then the same kernel size / stride / padding /
            dilation will be used along every spatial dimension.

            If they are a sequence then the sequence should be of length equal to
            `num_spatial_dims`, and specify the value of each property down each spatial
            dimension in turn.

            In addition, `padding` can be:

            - a sequence of 2-element tuples, each representing the padding to apply
                before and after each spatial dimension.
            - the string `'VALID'`, which is the same as zero padding.
            - one of the strings `'SAME'` or `'SAME_LOWER'`. This will apply padding to
                produce an output with the same size spatial dimensions as the input.
                The padding is split between the two sides equally or almost equally. In
                case the padding is an odd number, then the extra padding is added at
                the end for `'SAME'` and at the beginning for `'SAME_LOWER'`.
        """

        parse = _ntuple(num_spatial_dims)
        kernel_size = parse(kernel_size)
        stride = parse(stride)
        dilation = parse(dilation)

        if in_channels % groups != 0:
            raise ValueError(
                f"`in_channels` (={in_channels}) must be divisible "
                f"by `groups` (={groups})."
            )

        dtype = default_floating_dtype() if dtype is None else dtype
        wkey, bkey = jrandom.split(key, 2)
        grouped_in_channels = in_channels // groups
        lim = 1 / math.sqrt(grouped_in_channels * math.prod(kernel_size))
        wshape = (out_channels, grouped_in_channels) + kernel_size
        self.weight = default_init(wkey, wshape, dtype, lim)
        bshape = (out_channels,) + (1,) * num_spatial_dims
        self.bias = default_init(bkey, bshape, dtype, lim) if use_bias else None

        self.num_spatial_dims = num_spatial_dims
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = _padding_init(padding, num_spatial_dims)
        self.dilation = dilation
        self.groups = groups
        self.use_bias = use_bias
        self.padding_mode = _padding_mode_init(padding_mode)

    def _nonzero_pad(self, x: Array) -> Array:
        if isinstance(self.padding, str):
            rhs_shape = tuple(
                d * (k - 1) + 1 for k, d in zip(self.kernel_size, self.dilation)
            )
            padding = lax.padtype_to_pads(
                x.shape[1:], rhs_shape, self.stride, self.padding
            )
        else:
            padding = list(self.padding)

        if self.padding_mode == "REFLECT":
            mode = "reflect"
        elif self.padding_mode == "REPLICATE":
            mode = "edge"
        elif self.padding_mode == "CIRCULAR":
            mode = "wrap"
        else:
            raise ValueError("Invalid padding mode")

        x = jnp.pad(x, [(0, 0)] + padding, mode)
        return x

    @named_scope("eqx.nn.Conv")
    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape
            `(in_channels, dim_1, ..., dim_N)`, where `N = num_spatial_dims`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(out_channels, new_dim_1, ..., new_dim_N)`.
        """
        del key

        unbatched_rank = self.num_spatial_dims + 1
        if x.ndim != unbatched_rank:
            raise ValueError(
                f"Input to `Conv` needs to have rank {unbatched_rank},",
                f" but input has shape {x.shape}.",
            )

        if self.padding_mode != "ZEROS":
            x = self._nonzero_pad(x)
            padding = "VALID"
        else:
            padding = self.padding

        x = jnp.expand_dims(x, axis=0)
        x = lax.conv_general_dilated(
            lhs=x,
            rhs=self.weight,
            window_strides=self.stride,
            padding=padding,
            rhs_dilation=self.dilation,
            feature_group_count=self.groups,
        )
        x = jnp.squeeze(x, axis=0)

        if self.use_bias:
            x = x + cast(Array, self.bias)
        return x


class Conv1d(Conv):
    """As [`equinox.nn.Conv`][] with `num_spatial_dims=1`."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = 1,
        padding: str | int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        dilation: int | Sequence[int] = 1,
        groups: int = 1,
        use_bias: bool = True,
        padding_mode: str = "ZEROS",
        dtype=None,
        *,
        key: PRNGKeyArray,
    ):
        super().__init__(
            num_spatial_dims=1,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            use_bias=use_bias,
            padding_mode=padding_mode,
            dtype=dtype,
            key=key,
        )


class Conv2d(Conv):
    """As [`equinox.nn.Conv`][] with `num_spatial_dims=2`."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = (1, 1),
        padding: str | int | Sequence[int] | Sequence[tuple[int, int]] = (0, 0),
        dilation: int | Sequence[int] = (1, 1),
        groups: int = 1,
        use_bias: bool = True,
        padding_mode: str = "ZEROS",
        dtype=None,
        *,
        key: PRNGKeyArray,
    ):
        super().__init__(
            num_spatial_dims=2,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            use_bias=use_bias,
            padding_mode=padding_mode,
            dtype=dtype,
            key=key,
        )


class Conv3d(Conv):
    """As [`equinox.nn.Conv`][] with `num_spatial_dims=3`."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = (1, 1, 1),
        padding: str | int | Sequence[int] | Sequence[tuple[int, int]] = (0, 0, 0),
        dilation: int | Sequence[int] = (1, 1, 1),
        groups: int = 1,
        use_bias: bool = True,
        padding_mode: str = "ZEROS",
        dtype=None,
        *,
        key: PRNGKeyArray,
    ):
        super().__init__(
            num_spatial_dims=3,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            use_bias=use_bias,
            padding_mode=padding_mode,
            dtype=dtype,
            key=key,
        )


class ConvTranspose(Module):
    """General N-dimensional transposed convolution."""

    num_spatial_dims: int = field(static=True)
    weight: Array
    bias: Array | None
    in_channels: int = field(static=True)
    out_channels: int = field(static=True)
    kernel_size: tuple[int, ...] = field(static=True)
    stride: tuple[int, ...] = field(static=True)
    padding: str | tuple[tuple[int, int], ...] = field(static=True)
    output_padding: tuple[int, ...] = field(static=True)
    dilation: tuple[int, ...] = field(static=True)
    groups: int = field(static=True)
    use_bias: bool = field(static=True)
    padding_mode: str = field(static=True)

    def __init__(
        self,
        num_spatial_dims: int,
        in_channels: int,
        out_channels: int,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = 1,
        padding: str | int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        output_padding: int | Sequence[int] = 0,
        dilation: int | Sequence[int] = 1,
        groups: int = 1,
        use_bias: bool = True,
        padding_mode: str = "ZEROS",
        dtype=None,
        *,
        key: PRNGKeyArray,
    ):
        """**Arguments:**

        - `num_spatial_dims`: The number of spatial dimensions. For example traditional
            convolutions for image processing have this set to `2`.
        - `in_channels`: The number of input channels.
        - `out_channels`: The number of output channels.
        - `kernel_size`: The size of the transposed convolutional kernel.
        - `stride`: The stride used on the equivalent [`equinox.nn.Conv`][].
        - `padding`: The padding used on the equivalent [`equinox.nn.Conv`][].
        - `output_padding`: Additional padding for the output shape.
        - `dilation`: The spacing between kernel points.
        - `groups`: The number of input channel groups. At `groups=1`,
            all input channels contribute to all output channels. Values
            higher than 1 are equivalent to running `groups` independent
            `ConvTranspose` operations side-by-side, each having access only to
            `in_channels` // `groups` input channels, and
            concatenating the results along the output channel dimension.
            `in_channels` must be divisible by `groups`.
        - `use_bias`: Whether to add on a bias after the transposed convolution.
        - `padding_mode`: One of the following strings specifying the padding values
            used on the equivalent [`equinox.nn.Conv`][].
            - `'ZEROS'` (default): pads with zeros, no extra connectivity.
            - `'CIRCULAR'`: pads with circular values, extra connectivity (see the Tip
                below).
        - `dtype`: The dtype to use for the weight and the bias in this layer.
            Defaults to either `jax.numpy.float32` or `jax.numpy.float64` depending
            on whether JAX is in 64-bit mode.
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)

        !!! info

            All of `kernel_size`, `stride`, `padding`, `dilation` can be either an
            integer or a sequence of integers.

            If they are an integer then the same kernel size / stride / padding /
            dilation will be used along every spatial dimension.

            If they are a sequence then the sequence should be of length equal to
            `num_spatial_dims`, and specify the value of each property down each spatial
            dimension in turn.

            In addition, `padding` can be:

            - a sequence of 2-element tuples, each representing the padding to apply
                before and after each spatial dimension.
            - the string `'VALID'`, which is the same as zero padding.
            - one of the strings `'SAME'` or `'SAME_LOWER'`. This will apply padding to
                produce an output with the same size spatial dimensions as the input.
                The padding is split between the two sides equally or almost equally. In
                case the padding is an odd number, then the extra padding is added at
                the end for `'SAME'` and at the beginning for `'SAME_LOWER'`.

        !!! tip

            Transposed convolutions are often used to go in the "opposite direction" to
            a normal convolution. That is, from something with the shape of the output
            of a convolution to something with the shape of the input to a convolution.
            Moreover, to do so with the same "connectivity", i.e. which inputs can
            affect which outputs.

            Relative to an [`equinox.nn.Conv`][] layer, this can be accomplished by
            switching the values of `in_channels` and `out_channels`, whilst keeping
            `kernel_size`, `stride`, `padding`, `dilation`, and `groups` the same.

            When `stride > 1` then [`equinox.nn.Conv`][] maps multiple input shapes to
            the same output shape. `output_padding` is provided to resolve this
            ambiguity.

            - For `'SAME'` or `'SAME_LOWER'` padding, it reduces the calculated input
                shape.
            - For other cases, it adds a little extra padding to the bottom or right
                edges of the input.

            The extra connectivity created in 'CIRCULAR' padding is correctly taken into
            account. For instance, consider the equivalent
            [`equinox.nn.Conv`][] with kernel size 3. Then:

            - `Input 1234 --(zero padding)--> 012340 --(conv)--> Output abcd`
            - `Input 1234 --(circular padding)--> 412341 --(conv)--> Output abcd`

            so that `a` is connected with `1, 2` for zero padding, while connected with
            `1, 2, 4` for circular padding.

            See [these animations](https://github.com/vdumoulin/conv_arithmetic/blob/af6f818b0bb396c26da79899554682a8a499101d/README.md#transposed-convolution-animations)
            and [this report](https://arxiv.org/abs/1603.07285) for a nice reference.

        !!! faq "FAQ"

            If you need to exactly transpose a convolutional layer, i.e. not just create an
            operation with similar inductive biases but compute the actual linear transpose
            of a specific CNN you can reshape the weights of the forward convolution
            via the following:

            ```python
            cnn = eqx.Conv(...)
            cnn_t = eqx.ConvTranspose(...)
            cnn_t = eqx.tree_at(lambda x: x.weight, cnn_t, jnp.flip(cnn.weight,
                                axis=tuple(range(2, cnn.weight.ndim))).swapaxes(0, 1))
            ```

        !!! warning

            `padding_mode='CIRCULAR'` is only implemented for `output_padding=0` and
            `padding='SAME'` or `'SAME_LOWER'`.
        """  # noqa: E501
        dtype = default_floating_dtype() if dtype is None else dtype
        wkey, bkey = jrandom.split(key, 2)

        parse = _ntuple(num_spatial_dims)
        kernel_size = parse(kernel_size)
        stride = parse(stride)
        output_padding = parse(output_padding)
        dilation = parse(dilation)

        for s, o in zip(stride, output_padding, strict=True):
            if o >= s:
                raise ValueError("Must have `output_padding < stride` (elementwise).")

        grouped_in_channels = in_channels // groups
        lim = 1 / math.sqrt(grouped_in_channels * math.prod(kernel_size))
        wshape = (out_channels, grouped_in_channels) + kernel_size
        self.weight = default_init(wkey, wshape, dtype, lim)
        bshape = (out_channels,) + (1,) * num_spatial_dims
        self.bias = default_init(bkey, bshape, dtype, lim) if use_bias else None

        padding = _padding_init(padding, num_spatial_dims)
        padding_mode = _padding_mode_init(padding_mode)
        if padding_mode in ("REFLECT", "REPLICATE"):
            raise NotImplementedError(
                "'REFLECT' or 'REPLICATE' padding mode is not implemented"
            )
        elif padding_mode == "CIRCULAR":
            if any(o != 0 for o in output_padding):
                raise NotImplementedError(
                    "`padding_mode == 'CIRCULAR'` with non-zero `output_padding` is not"
                    "implemented."
                )
            if padding not in ("SAME", "SAME_LOWER"):
                raise NotImplementedError(
                    "`padding_mode == 'CIRCULAR'` is only implemented for 'SAME' or"
                    "'SAME_LOWER' padding."
                )

        self.num_spatial_dims = num_spatial_dims
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.output_padding = output_padding
        self.dilation = dilation
        self.groups = groups
        self.use_bias = use_bias
        self.padding_mode = padding_mode

    def _padding_transpose(self) -> tuple[tuple[int, int], ...]:
        # Notations follow https://arxiv.org/abs/1603.07285
        k = np.asarray(self.kernel_size)
        s = np.asarray(self.stride)
        a = np.asarray(self.output_padding)
        d = np.asarray(self.dilation)

        if isinstance(self.padding, str):
            if self.padding == "VALID":
                p0 = np.zeros(self.num_spatial_dims, np.int64)
                p1 = np.zeros(self.num_spatial_dims, np.int64)
            else:
                p_sum = d * (k - 1) - s + a + 1
                a = np.where(p_sum < 0, -p_sum, 0)
                p_sum = np.where(p_sum < 0, 0, p_sum)
                lower = 1 if self.padding == "SAME_LOWER" else 0
                p0 = (p_sum + lower) // 2
                p1 = p_sum - p0
        else:
            p0 = np.asarray(self.padding)[:, 0]
            p1 = np.asarray(self.padding)[:, 1]

        # Given by Relationship 14 of https://arxiv.org/abs/1603.07285
        p0t = d * (k - 1) - p0
        p1t = d * (k - 1) - p1 + a
        padding_t = tuple((x.item(), y.item()) for x, y in zip(p0t, p1t))
        return padding_t

    def _circular_pad(
        self, x: Array, padding_t: tuple[tuple[int, int], ...]
    ) -> tuple[Array, tuple[tuple[int, int], ...]]:
        stride = np.expand_dims(self.stride, axis=1)
        pad_width = np.insert(padding_t // stride, 0, 0, axis=0)
        x = jnp.pad(x, pad_width, mode="wrap")
        padding_t = tuple((p[0].item(), p[1].item()) for p in padding_t % stride)  # pyright: ignore[reportIndexIssue]
        return x, padding_t

    @named_scope("eqx.nn.ConvTranspose")
    def __call__(self, x: Array, *, key: PRNGKeyArray | None = None) -> Array:
        """**Arguments:**

        - `x`: The input. Should be a JAX array of shape
            `(in_channels, dim_1, ..., dim_N)`, where `N = num_spatial_dims`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        A JAX array of shape `(out_channels, new_dim_1, ..., new_dim_N)`.
        """
        del key
        unbatched_rank = self.num_spatial_dims + 1
        if x.ndim != unbatched_rank:
            raise ValueError(
                f"Input to `ConvTranspose` needs to have rank {unbatched_rank},",
                f" but input has shape {x.shape}.",
            )

        padding_t = self._padding_transpose()
        if self.padding_mode == "CIRCULAR":
            x, padding_t = self._circular_pad(x, padding_t)

        x = jnp.expand_dims(x, axis=0)
        x = lax.conv_general_dilated(
            lhs=x,
            rhs=self.weight,
            window_strides=(1,) * self.num_spatial_dims,
            padding=padding_t,
            lhs_dilation=self.stride,
            rhs_dilation=self.dilation,
            feature_group_count=self.groups,
        )
        x = jnp.squeeze(x, axis=0)

        if self.use_bias:
            x = x + cast(Array, self.bias)
        return x


class ConvTranspose1d(ConvTranspose):
    """As [`equinox.nn.ConvTranspose`][] with `num_spatial_dims=1`."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = 1,
        output_padding: int | Sequence[int] = 0,
        padding: str | int | Sequence[int] | Sequence[tuple[int, int]] = 0,
        dilation: int | Sequence[int] = 1,
        groups: int = 1,
        use_bias: bool = True,
        padding_mode: str = "ZEROS",
        dtype=None,
        *,
        key: PRNGKeyArray,
    ):
        super().__init__(
            num_spatial_dims=1,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            output_padding=output_padding,
            padding=padding,
            dilation=dilation,
            groups=groups,
            use_bias=use_bias,
            padding_mode=padding_mode,
            dtype=dtype,
            key=key,
        )


class ConvTranspose2d(ConvTranspose):
    """As [`equinox.nn.ConvTranspose`][] with `num_spatial_dims=2`."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = (1, 1),
        output_padding: int | Sequence[int] = (0, 0),
        padding: str | int | Sequence[int] | Sequence[tuple[int, int]] = (0, 0),
        dilation: int | Sequence[int] = (1, 1),
        groups: int = 1,
        use_bias: bool = True,
        padding_mode: str = "ZEROS",
        dtype=None,
        *,
        key: PRNGKeyArray,
    ):
        super().__init__(
            num_spatial_dims=2,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            output_padding=output_padding,
            padding=padding,
            dilation=dilation,
            groups=groups,
            use_bias=use_bias,
            padding_mode=padding_mode,
            dtype=dtype,
            key=key,
        )


class ConvTranspose3d(ConvTranspose):
    """As [`equinox.nn.ConvTranspose`][] with `num_spatial_dims=3`."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int | Sequence[int],
        stride: int | Sequence[int] = (1, 1, 1),
        output_padding: int | Sequence[int] = (0, 0, 0),
        padding: str | int | Sequence[int] | Sequence[tuple[int, int]] = (0, 0, 0),
        dilation: int | Sequence[int] = (1, 1, 1),
        groups: int = 1,
        use_bias: bool = True,
        padding_mode: str = "ZEROS",
        dtype=None,
        *,
        key: PRNGKeyArray,
    ):
        super().__init__(
            num_spatial_dims=3,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            output_padding=output_padding,
            padding=padding,
            dilation=dilation,
            groups=groups,
            use_bias=use_bias,
            padding_mode=padding_mode,
            dtype=dtype,
            key=key,
        )



"""This module provides operations to assert that a particular value is always
unbatched, or not differentiated, etc.
"""

import functools as ft

import jax
import jax.extend.core
import jax.interpreters.ad as ad
import jax.interpreters.batching as batching
import jax.interpreters.mlir as mlir
import jax.numpy as jnp
import jax.tree_util as jtu
from jaxtyping import PyTree

from .._ad import nondifferentiable as nondifferentiable  # public re-export
from .._errors import error_if
from .._filters import combine, is_array, partition


_nontraceable_impl = lambda x, *, name: x


def _make_error(opname):
    def _error(*args, name):
        raise RuntimeError(f"Detected {opname} of {name}")

    return _error


nontraceable_p = jax.extend.core.Primitive("nontraceable")
nontraceable_p.def_impl(_nontraceable_impl)
nontraceable_p.def_abstract_eval(_nontraceable_impl)
ad.primitive_jvps[nontraceable_p] = _make_error("differentiation")
ad.primitive_transposes[nontraceable_p] = _make_error("transposition")
batching.primitive_batchers[nontraceable_p] = _make_error("batching")
mlir.register_lowering(
    nontraceable_p, mlir.lower_fun(_nontraceable_impl, multiple_results=False)
)


def nontraceable(x, *, name="nontraceable operation"):
    """Identity function, which raises an error if it is transformed in any way. (i.e.
    in `jax.grad`, `jax.vmap` etc.)

    This is useful at the end of the `impl` rule for higher-order final-style
    primitives, for checking that no other tracers were captured via closure.
    """
    dynamic, static = partition(x, is_array)
    bind = ft.partial(nontraceable_p.bind, name=name)
    dynamic = jtu.tree_map(bind, dynamic)
    return combine(dynamic, static)


nondifferentiable_backward_p = jax.extend.core.Primitive("nondifferentiable_backward")


def _nondifferentiable_backward_batch(x, batch_axes, *, msg, symbolic):
    (x,) = x
    (batch_axes,) = batch_axes
    return nondifferentiable_backward(x, msg=msg, symbolic=symbolic), batch_axes


def _nondifferentiable_backward_jvp(primals, tangents, *, msg, symbolic):
    (primals,) = primals
    (tangents,) = tangents
    primal_out = nondifferentiable_backward(primals, msg=msg, symbolic=symbolic)
    tangent_out = nondifferentiable_backward(tangents, msg=msg, symbolic=symbolic)
    return primal_out, tangent_out


def _nondifferentiable_backward_transpose(cts_in, _, *, msg, symbolic):
    if isinstance(cts_in, ad.Zero):
        return ad.Zero  # the class, not an instance
    else:
        if symbolic:
            raise RuntimeError(msg)
        else:
            # Unfortunately there are legitimate cases where we get all-zero
            # non-symbolic cotangents, so we have to use a runtime error here instead of
            # just erroring at trace time.
            # This happens when doing something like:
            #
            # x = nondifferentiable_backward(x)
            # x, y = lax.cond(pred, lambda: (x, y), lambda: (x+1, y+1))
            # return y
            return [error_if(cts_in, (cts_in != 0).any(), msg)]


_nondifferentiable_backward_impl = lambda x, *, msg, symbolic: x


nondifferentiable_backward_p.def_impl(_nondifferentiable_backward_impl)
nondifferentiable_backward_p.def_abstract_eval(_nondifferentiable_backward_impl)
ad.primitive_jvps[nondifferentiable_backward_p] = _nondifferentiable_backward_jvp
ad.primitive_transposes[nondifferentiable_backward_p] = (
    _nondifferentiable_backward_transpose
)
batching.primitive_batchers[nondifferentiable_backward_p] = (
    _nondifferentiable_backward_batch
)
mlir.register_lowering(
    nondifferentiable_backward_p,
    mlir.lower_fun(_nondifferentiable_backward_impl, multiple_results=False),
)


def nondifferentiable_backward(
    x: PyTree,
    name: str | None = None,
    msg: str | None = None,
    symbolic: bool = True,
) -> PyTree:
    """Identity function. Raises an error if it is differentiated in reverse mode."""
    dynamic, static = partition(x, is_array)
    flat, treedef = jtu.tree_flatten(dynamic)
    if msg is None:
        if name is None:
            name = "This operation"
        msg = f"Unexpected cotangent. {name} cannot be reverse-mode autodifferentiated."
    bind = ft.partial(nondifferentiable_backward_p.bind, msg=msg, symbolic=symbolic)
    flat = map(bind, flat)
    return combine(jtu.tree_unflatten(treedef, flat), static)


def _cannot_batch(x, b, *, msg, allow_constant_across_batch):
    (x,) = x
    (b,) = b
    if b is None:
        return x, b
    else:
        if allow_constant_across_batch:
            x = error_if(x, jnp.min(x, axis=b) != jnp.max(x, axis=b), msg)
            return jnp.take(x, 0, axis=b), None
        else:
            raise ValueError(msg)


nonbatchable_p = jax.extend.core.Primitive("nonbatchable")
nonbatchable_p.def_impl(lambda x, *, msg, allow_constant_across_batch: x)
nonbatchable_p.def_abstract_eval(lambda x, *, msg, allow_constant_across_batch: x)
batching.primitive_batchers[nonbatchable_p] = _cannot_batch
mlir.register_lowering(
    nonbatchable_p,
    mlir.lower_fun(
        lambda x, *, msg, allow_constant_across_batch: x, multiple_results=False
    ),
)


def nonbatchable(
    x: PyTree,
    *,
    name: str | None = None,
    msg: str | None = None,
    allow_constant_across_batch: bool = False,
) -> PyTree:
    """Identity function. Raises a trace-time assert if it is batched."""
    dynamic, static = partition(x, is_array)
    flat, treedef = jtu.tree_flatten(dynamic)
    if msg is None:
        if name is None:
            name = "This operation"
        if allow_constant_across_batch:
            msg = (
                f"Nonconstant batch. {name} has received a batch of values that were "
                "expected to be constant. This is probably an internal error in the "
                "library you are using."
            )
        else:
            msg = (
                f"Unexpected batch tracer. {name} cannot be vmap'd. This is probably "
                "an internal error in the library you are using."
            )
    bind = ft.partial(
        nonbatchable_p.bind,
        msg=msg,
        allow_constant_across_batch=allow_constant_across_batch,
    )
    flat = map(bind, flat)
    return combine(jtu.tree_unflatten(treedef, flat), static)



import math
from typing import cast

import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jrandom
from jaxtyping import Array, PRNGKeyArray

from .._misc import default_floating_dtype
from .._module import field, Module
from ._misc import default_init, named_scope


class GRUCell(Module):
    """A single step of a Gated Recurrent Unit (GRU).

    !!! example

        This is often used by wrapping it into a `jax.lax.scan`. For example:

        ```python
        class Model(Module):
            cell: GRUCell

            def __init__(self, **kwargs):
                self.cell = GRUCell(**kwargs)

            def __call__(self, xs):
                scan_fn = lambda state, input: (self.cell(input, state), None)
                init_state = jnp.zeros(self.cell.hidden_size)
                final_state, _ = jax.lax.scan(scan_fn, init_state, xs)
                return final_state
        ```
    """

    weight_ih: Array
    weight_hh: Array
    bias: Array | None
    bias_n: Array | None
    input_size: int = field(static=True)
    hidden_size: int = field(static=True)
    use_bias: bool = field(static=True)

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        use_bias: bool = True,
        dtype=None,
        *,
        key: PRNGKeyArray,
    ):
        """**Arguments:**

        - `input_size`: The dimensionality of the input vector at each time step.
        - `hidden_size`: The dimensionality of the hidden state passed along between
            time steps.
        - `use_bias`: Whether to add on a bias after each update.
        - `dtype`: The dtype to use for all weights and biases in this GRU cell.
            Defaults to either `jax.numpy.float32` or `jax.numpy.float64` depending on
            whether JAX is in 64-bit mode.
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)
        """
        dtype = default_floating_dtype() if dtype is None else dtype
        ihkey, hhkey, bkey, bkey2 = jrandom.split(key, 4)
        lim = math.sqrt(1 / hidden_size)

        ihshape = (3 * hidden_size, input_size)
        self.weight_ih = default_init(ihkey, ihshape, dtype, lim)
        hhshape = (3 * hidden_size, hidden_size)
        self.weight_hh = default_init(hhkey, hhshape, dtype, lim)
        if use_bias:
            self.bias = default_init(bkey, (3 * hidden_size,), dtype, lim)
            self.bias_n = default_init(bkey2, (hidden_size,), dtype, lim)
        else:
            self.bias = None
            self.bias_n = None

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.use_bias = use_bias

    @named_scope("eqx.nn.GRUCell")
    def __call__(self, input: Array, hidden: Array, *, key: PRNGKeyArray | None = None):
        """**Arguments:**

        - `input`: The input, which should be a JAX array of shape `(input_size,)`.
        - `hidden`: The hidden state, which should be a JAX array of shape
            `(hidden_size,)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        The updated hidden state, which is a JAX array of shape `(hidden_size,)`.
        """
        if self.use_bias:
            bias = cast(Array, self.bias)
            bias_n = cast(Array, self.bias_n)
        else:
            bias = 0
            bias_n = 0
        igates = jnp.split(self.weight_ih @ input + bias, 3)
        hgates = jnp.split(self.weight_hh @ hidden, 3)
        reset = jnn.sigmoid(igates[0] + hgates[0])
        inp = jnn.sigmoid(igates[1] + hgates[1])
        new = jnn.tanh(igates[2] + reset * (hgates[2] + bias_n))
        return new + inp * (hidden - new)


class LSTMCell(Module):
    """A single step of a Long-Short Term Memory unit (LSTM).

    !!! example

        This is often used by wrapping it into a `jax.lax.scan`. For example:

        ```python
        class Model(Module):
            cell: LSTMCell

            def __init__(self, ...):
                self.cell = LSTMCell(...)

            def __call__(self, xs):
                scan_fn = lambda state, input: (self.cell(input, state), None)
                init_state = (jnp.zeros(self.cell.hidden_size),
                              jnp.zeros(self.cell.hidden_size))
                (h, c), _ = jax.lax.scan(scan_fn, init_state, xs)
                return h, c
        ```
    """

    weight_ih: Array
    weight_hh: Array
    bias: Array | None
    input_size: int = field(static=True)
    hidden_size: int = field(static=True)
    use_bias: bool = field(static=True)

    def __init__(
        self,
        input_size: int,
        hidden_size: int,
        use_bias: bool = True,
        dtype=None,
        *,
        key: PRNGKeyArray,
    ):
        """**Arguments:**

        - `input_size`: The dimensionality of the input vector at each time step.
        - `hidden_size`: The dimensionality of the hidden state passed along between
            time steps.
        - `use_bias`: Whether to add on a bias after each update.
        - `dtype`: The dtype to use for all weights and biases in this LSTM cell.
            Defaults to either `jax.numpy.float32` or `jax.numpy.float64` depending on
            whether JAX is in 64-bit mode.
        - `key`: A `jax.random.PRNGKey` used to provide randomness for parameter
            initialisation. (Keyword only argument.)
        """
        dtype = default_floating_dtype() if dtype is None else dtype
        ihkey, hhkey, bkey = jrandom.split(key, 3)
        lim = math.sqrt(1 / hidden_size)

        ihshape = (4 * hidden_size, input_size)
        self.weight_ih = default_init(ihkey, ihshape, dtype, lim)
        hhshape = (4 * hidden_size, hidden_size)
        self.weight_hh = default_init(hhkey, hhshape, dtype, lim)
        bshape = (4 * hidden_size,)
        self.bias = default_init(bkey, bshape, dtype, lim) if use_bias else None

        self.input_size = input_size
        self.hidden_size = hidden_size
        self.use_bias = use_bias

    @named_scope("eqx.nn.LSTMCell")
    def __call__(self, input, hidden, *, key=None):
        """**Arguments:**

        - `input`: The input, which should be a JAX array of shape `(input_size,)`.
        - `hidden`: The hidden state, which should be a 2-tuple of JAX arrays, each of
            shape `(hidden_size,)`.
        - `key`: Ignored; provided for compatibility with the rest of the Equinox API.
            (Keyword only argument.)

        **Returns:**

        The updated hidden state, which is a 2-tuple of JAX arrays, each of shape
        `(hidden_size,)`.
        """
        h, c = hidden
        lin = self.weight_ih @ input + self.weight_hh @ h
        if self.use_bias:
            lin = lin + self.bias
        i, f, g, o = jnp.split(lin, 4)
        i = jnn.sigmoid(i)
        f = jnn.sigmoid(f)
        g = jnn.tanh(g)
        o = jnn.sigmoid(o)
        c = f * c + i * g
        h = o * jnn.tanh(c)
        return (h, c)
